# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

"""create_solar_forecast
This module contains the CRON job that is periodically executed to make
prognoses of solar features that are usefull for splitting the load in solar and
wind contributions.
Example:
    This module is meant to be called directly from a CRON job. A description of
    the CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::
        $ python create_solar_forecast
Attributes:

"""
from datetime import datetime

import numpy as np

from openstf.model.general import fides
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext


def make_solar_predicion_pj(pj, context):
    """Make a solar prediction for a spcecific prediction job.

    Args:
        pj: (dict) prediction job
    """
    context.logger.info("Get solar input data from database")
    # pvdata is only stored in the prd database
    solar_input = context.database.get_solar_input(
        (pj["lat"], pj["lon"]),
        pj["horizon_minutes"],
        pj["resolution_minutes"],
        radius=pj["radius"],
        sid=pj["sid"],
    )

    if len(solar_input) == 0:
        raise ValueError("Empty solar input")

    context.logger.info("Make solar prediction using Fides")
    power = fides(
        solar_input[["aggregated", "radiation"]].rename(
            columns=dict(radiation="insolation", aggregated="load")
        )
    )

    # if the forecast is for a region, output should be scaled to peak power
    if (pj["radius"] != 0) and (not np.isnan(pj["peak_power"])):
        power = pj["peak_power"] / max(solar_input.aggregated) * power
    context.logger.info("Store solar prediction in database")
    power["pid"] = pj["id"]
    power["type"] = "solar"
    power["algtype"] = "Fides"
    power["customer"] = pj["name"]
    power["description"] = pj["description"]
    context.database.write_forecast(power)


def main():
    with TaskContext(__file__) as context:
        context.logger.info("Querying wind prediction jobs from database")
        prediction_jobs = context.database.get_prediction_jobs_solar()
        num_prediction_jobs = len(prediction_jobs)

        # only make customer = Provincie once an hour
        utc_now_minute = datetime.utcnow().minute
        if utc_now_minute >= 15:
            prediction_jobs = [
                pj for pj in prediction_jobs if str(pj["name"]).startswith("Provincie")
            ]
            num_removed_jobs = num_prediction_jobs - len(prediction_jobs)
            num_prediction_jobs = len(prediction_jobs)
            context.logger.info(
                "Remove 'Provincie' solar predictions",
                num_removed_jobs=num_removed_jobs,
                num_prediction_jobs=num_prediction_jobs,
            )

        PredictionJobLoop(context, prediction_jobs=prediction_jobs).map(
            make_solar_predicion_pj, context
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import pickle

import lightgbm as lgb

from openstf.model.serializer.serializer import AbstractModelSerializer


class LGBModelSerializer(AbstractModelSerializer):
    def __init__(self):
        super().__init__()

    def save(self, pid, lgb_model, corrections=None):

        save_folder = self.build_save_folder_path(pid=pid)

        save_folder.mkdir(parents=True, exist_ok=True)

        # save LGB model
        self.logger.warning(f"Save folder: {save_folder}")
        lgb_model.save_model(str(save_folder / self.MODEL_FILENAME))

        # save corrections
        if corrections is not None:
            if len(corrections) == len(corrections.dropna()):
                corrections.to_csv(
                    save_folder / self.CORRECTIONS_FILENAME, **self.PANDAS_CSV_FORMAT
                )
            return

        self.logger.warning(
            "No corrections found, corrections file not saved for pid: {pid}"
        )

    # NOTE from general.py
    def load(self, pid, pid_model_folder=None):
        """Load the most recent model.

        Args:
            pid (int): Prediction job id
            pid_model_folder(str): Path to save model at specific location

        Raises:
            FileNotFoundError: If no recent model file was found

        Returns:
            tuple: Tuple with:
                [0]: lgb.Booster: Loaded model
                [1]: str: Path to loaded model
        """

        if pid_model_folder is None:
            pid_model_folder = self.build_pid_model_folder_path(pid=pid)

        try:
            model_file, model_folder = self.find_most_recent_model_file(
                pid_model_folder
            )
        except FileNotFoundError as e:
            self.logger.error(f"Can't load model, no recent model file found. {e}")
            return None, None
        model_file = str(model_file)
        lgb_model = lgb.Booster(model_file=model_file)
        self.logger.info(f"Loaded model from: {model_file}")

        return lgb_model, model_file



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime

import numpy as np
import pandas as pd
from ktpbase.database import DataBase
from ktpbase.log import logging
import scipy.optimize

import openstf.monitoring.teams as monitoring

COEF_MAX_FRACTION_DIFF = 0.3


def split_energy(pid):
    """Function that caries out the energy splitting for a specific prediction job with
    id pid.

    Args:
        pid (int): Prediction job id

    Returns:
        pandas.DataFrame: Energy splitting coefficients.
    """
    # Make database connection
    db = DataBase()
    logger = logging.get_logger(__name__)

    # Get Prediction job
    pj = db.get_prediction_job(pid)

    logger.info("Start splitting energy", pid=pj["id"])

    # Get input for splitting
    input_split_function = db.get_input_energy_splitting(pj)

    # Carry out the splitting
    components, coefdict = find_components(input_split_function)

    # Calculate mean absolute error (MAE)
    # TODO: use a standard metric function for this
    error = components[["load", "Inschatting"]].diff(axis=1).iloc[:, 1]
    mae = error.abs().mean()
    coefdict.update({"MAE": mae})
    coefsdf = convert_coefdict_to_coefsdf(pj, input_split_function, coefdict)

    # Get the coefs of previous runs and check if new coefs are valid
    last_coefsdict = db.get_energy_split_coefs(pj)
    last_coefsdf = convert_coefdict_to_coefsdf(pj, input_split_function, last_coefsdict)
    invalid_coefs = determine_invalid_coefs(coefsdf, last_coefsdf)
    if not invalid_coefs.empty:
        # If coefs not valid, do not update the coefs in the db and send teams
        # message that something strange is happening
        monitoring.post_teams(
            f"New splitting coefficient(s) for pid **{pj['id']}** deviate strongly "
            f"from previously stored coefficients.",
            invalid_coefs=invalid_coefs,
            coefsdf=coefsdf,
        )
        # Use the last known coefficients for further processing
        return last_coefsdf
    else:
        # Save Results
        db.write_energy_splitting_coefficients(coefsdf, if_exists="append")
        logger.info(
            "Succesfully wrote energy split coefficients to database", pid=pj["id"]
        )
        return coefsdf


def determine_invalid_coefs(new_coefs, last_coefs):
    """Determine which new coefficients are valid and return them.

    Args:
        new_coefs (pd.DataFrame): df of new coefficients for standard load
            profiles (i.e. wind, solar, household)
        last_coefs (pd.DataFrame): df of last coefficients for standard load
            profiles (i.e. wind, solar, household)

    Returns:
        pd.DataFrame: df of invalid coefficients
    """
    merged_coefs = pd.merge(
        last_coefs, new_coefs, on="coef_name", how="left", suffixes=["_last", "_new"]
    )
    # calculate difference between new and last coefficients, if no new
    # coefficient, set difference to inf
    # If coefficient name is not present in new coefficients list, fail. If coefficient
    # name is not present in last coefficients list, add it.
    merged_coefs["difference"] = (
        (merged_coefs.coef_value_last - merged_coefs.coef_value_new)
        .abs()
        .fillna(np.inf)
    )
    # Check if the absolute difference between last coefficients and new coefficients
    # is more than COEF_MAX_FRACTION_DIFF x absolute value of last coefficient
    invalid_coefs = merged_coefs[
        merged_coefs.difference
        > (COEF_MAX_FRACTION_DIFF * merged_coefs.coef_value_last).abs()
    ]
    return invalid_coefs


def convert_coefdict_to_coefsdf(pj, input_split_function, coefdict):
    """Convert dictionary of coefficients to dataframe with additional data for db
    storage.

    Args:
        pj (PredictionJob): prediction job
        input_split_function (pd.DataFrame): df of columns of standard load profiles,
            i.e. wind, solar, household
        coefdict (dict): dict of coefficient per standard load profile

    Returns:
        pd.DataFrame: df of coefficients to insert in sql
    """
    #
    sql_column_labels = ["pid", "date_start", "date_end", "created"]
    sql_colum_values = [
        pj["id"],
        input_split_function.index.min().date(),
        input_split_function.index.max().date(),
        datetime.utcnow(),
    ]
    coefsdf = pd.DataFrame(
        {"coef_name": list(coefdict.keys()), "coef_value": list(coefdict.values())}
    )
    for i, column in enumerate(sql_column_labels):
        coefsdf[column] = sql_colum_values[i]

    return coefsdf


def find_components(df, zero_bound=True):
    """Function that does the actual energy splitting

    Args:
        df (pandas.DataFrame): Input data. The dataframe should contain these columns
            in exactly this order: [load, wind_ref, pv_ref, mulitple tdcv colums]
        zero_bound (bool): If zero_bound is True coefficients can't be negative.

    Returns:
        tuple:
            [0] pandas.DataFrame: Containing the wind and solar components
            [1] dict: The coefficients that result from the fitting
    """

    # Define function to fit
    def weighted_sum(x, *args):
        if len(x) != len(args):
            raise Exception("Length of args should match len of x")
        weights = np.array([v for v in args])
        return np.dot(x.T, weights)

    load = df.iloc[:, 0]
    wind_ref = df.iloc[:, 1]
    pv_ref = df.iloc[:, 2]

    # Define scaler
    nedu_scaler = (load.max() - load.min()) / 10

    # Come up with inital guess for the fitting
    p_wind_guess = 1.0
    ppv_guess = 1.0
    p0 = [p_wind_guess, ppv_guess] + (len(df.columns) - 3) * [nedu_scaler]

    # Define fitting bounds
    if zero_bound:
        bounds = (0, "inf")
    else:
        bounds = ("-inf", "inf")

    # Carry out fitting
    # See https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html # noqa
    coefs, cov = scipy.optimize.curve_fit(
        weighted_sum,
        xdata=df.iloc[:, 1:].values.T,
        ydata=load.values,
        p0=p0,
        bounds=bounds,
        method="trf",
    )

    # Set 'almost zero' to zero
    coefs[coefs < 0.1] = 0

    # Reconstuct historical load
    hist = weighted_sum(df.iloc[:, 1:].values.T, *coefs)
    histp0 = weighted_sum(df.iloc[:, 1:].values.T, *p0)

    # Make a nice dataframe to return the components
    components = df.iloc[:, [0]].copy()
    components["Inschatting"] = hist.T
    components["p0"] = histp0.T
    components["Windopwek"] = wind_ref * coefs[0]
    components["Zonne-opwek"] = pv_ref * coefs[1]
    components["StandaardVerbruik"] = (df.iloc[:, 3:] * coefs[2:]).sum(axis=1)
    components["Residu"] = -1 * components.iloc[:, 0:2].diff(axis=1).iloc[:, 1]

    # Make nice dictinary to return coefficents
    coefdict = {name: value for name, value in zip(df.columns[1:], coefs)}

    # Return result
    return components, coefdict



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from ktpbase.database import DataBase
from ktpbase.log import logging

from openstf.feature_engineering.general import calc_completeness
from openstf.model import metrics


def calc_kpi_for_specific_pid(pid, start_time=None, end_time=None):
    """Function that checks the model performance based on a pid. This function
    - loads and combines forecast and realised data
    - calculated several key performance indicators (KPIs)
    These metric include:
        - RMSE,
        - bias,
        - NSME (model efficiency, between -inf and 1)
        - Mean absolute Error

    Args:
        pid (int): Prediction ID for a given prediction job
        start_time (datetime): Start time from when to retrieve the historic load prediction.
        end_time (datetime): Start time till when to retrieve the historic load prediction.

    Returns:
        Dictionary that includes a dictonary for each t_ahead.
        Dict includes enddate en window (in days) for clarification

    Example:
        To get the rMAE for the 24 hours ahead prediction: kpis['24h']['rMAE']
    """
    COMPLETENESS_REALISED_THRESHOLDS = 0.7
    COMPLETENESS_PREDICTED_LOAD_THRESHOLD = 0.7

    # Make database connection
    db = DataBase()
    log = logging.get_logger(__name__)

    pj = db.get_prediction_job(pid)

    # Apply default parameters if none are provided
    if start_time is None:
        start_time = datetime.utcnow() - timedelta(days=1)
    if end_time is None:
        end_time = datetime.utcnow()

    # Get realised load data
    realised = db.get_load_pid(pj["id"], start_time, end_time, "15T")

    # Get predicted load
    predicted_load = db.get_predicted_load_tahead(pj, start_time, end_time)

    # If predicted is empty, exit
    if len(predicted_load) == 0:
        raise ValueError(
            "Found no predicted load for pid {} between {} and {}!".format(
                pj["id"], start_time, end_time
            )
        )

    # If realised is empty, exit
    if len(realised) == 0:
        raise ValueError(
            "Found no realised load for pid {} between {} and {}!".format(
                pj["id"], start_time, end_time
            )
        )

    # Calculate completeness en raise exception if completeness does not meet requirements
    completeness_realised = calc_completeness(realised)

    # Interpolate missing data if needed
    realised = realised.resample("15T").interpolate(limit=3)

    # Calculate completeness for realised load
    completeness_predicted_load = calc_completeness(predicted_load)

    # Combine the forecast and the realised to make sure indices are matched nicely
    combined = pd.merge(realised, predicted_load, left_index=True, right_index=True)

    # TODO: make basecase calculation optionalm only if there was data available 7 days before
    # Add basecase (load in same time period 7 days ago)
    basecase = db.get_load_pid(
        pj["id"], start_time - timedelta(days=7), end_time - timedelta(days=7), "15T"
    ).shift(periods=7, freq="d")
    basecase = basecase.rename(columns=dict(load="basecase"))

    combined = combined.merge(basecase, how="left", left_index=True, right_index=True)

    # Raise exception in case of constant load
    if combined.load.nunique() == 1:
        raise Exception("The load is constant!")

    # Define output dictonary
    kpis = dict()

    # Extract t_aheads from predicted_load,
    # Make a list of tuples with [(forecast_xh, stdev_xh),(..,..),..]
    hor_list = [
        ("forecast_" + t_ahead, "stdev_" + t_ahead)
        for t_ahead in set(col.split("_")[1] for col in predicted_load.columns)
    ]

    # cast date to int
    date = pd.to_datetime(end_time)

    # Calculate model metrics and add them to the output dictionary
    log.info("Start calculating kpis")
    for hor_cols in hor_list:
        t_ahead_h = hor_cols[0].split("_")[1]
        fc = combined[hor_cols[0]]  # load predictions
        st = combined[hor_cols[1]]  # standard deviations of load predictions
        completeness_predicted_load = calc_completeness(fc.to_frame(name=t_ahead_h))
        kpis.update(
            {
                t_ahead_h: {
                    "RMSE": metrics.rmse(combined["load"], fc),
                    "bias": metrics.bias(combined["load"], fc),
                    "NSME": metrics.nsme(combined["load"], fc),
                    "MAE": metrics.mae(combined["load"], fc),
                    "rMAE": metrics.r_mae(combined["load"], fc),
                    "rMAE_highest": metrics.r_mae_highest(combined["load"], fc),
                    "rMNE_highest": metrics.r_mne_highest(combined["load"], fc),
                    "rMPE_highest": metrics.r_mpe_highest(combined["load"], fc),
                    "rMAE_lowest": metrics.r_mae_lowest(combined["load"], fc),
                    "skill_score_basecase": metrics.skill_score(
                        combined["load"],
                        combined["basecase"],
                        np.mean(combined["basecase"]),
                    ),
                    "skill_score": metrics.skill_score(
                        combined["load"], fc, np.mean(combined["basecase"])
                    ),
                    "skill_score_positive_peaks": metrics.skill_score_positive_peaks(
                        combined["load"], fc, np.mean(combined["basecase"])
                    ),
                    "skill_score_positive_peaks_basecase": metrics.skill_score_positive_peaks(
                        combined["load"],
                        combined["basecase"],
                        np.mean(combined["basecase"]),
                    ),
                    "franks_skill_score": metrics.franks_skill_score(
                        combined["load"], fc, combined["basecase"]
                    ),
                    "franks_skill_score_peaks": metrics.franks_skill_score_peaks(
                        combined["load"], fc, combined["basecase"]
                    ),
                    "load_range": combined["load"].max() - combined["load"].min(),
                    "frac_in_1sdev": metrics.frac_in_stdev(combined["load"], fc, st),
                    "frac_in_2sdev": metrics.frac_in_stdev(
                        combined["load"], fc, 2 * st
                    ),
                    "completeness_realised": completeness_realised,
                    "completeness_predicted": completeness_predicted_load,
                    "date": date,  # cast to date
                    "window_days": np.round(
                        (end_time - start_time).total_seconds() / 60.0 / 60.0 / 24.0
                    ),
                }
            }
        )

        if completeness_realised < COMPLETENESS_REALISED_THRESHOLDS:
            log.warning(
                "Completeness realised load too low",
                prediction_id=pj["id"],
                start_time=start_time,
                end_time=end_time,
                completeness=completeness_realised,
                completeness_threshold=COMPLETENESS_REALISED_THRESHOLDS,
            )
            set_incomplete_kpi_to_nan(kpis, t_ahead_h)
        if completeness_predicted_load < COMPLETENESS_PREDICTED_LOAD_THRESHOLD:
            log.warning(
                "Completeness predicted load too low",
                prediction_id=pj["id"],
                horizon=t_ahead_h,
                start_time=start_time,
                end_time=end_time,
                completeness=completeness_predicted_load,
                completeness_threshold=COMPLETENESS_PREDICTED_LOAD_THRESHOLD,
            )
            set_incomplete_kpi_to_nan(kpis, t_ahead_h)

    # Return output dictionary
    return kpis


def set_incomplete_kpi_to_nan(kpis, t_ahead_h):
    """
    Checks the given kpis for completeness and sets to nan if this not true

    :param kpis: the kpis
    :param t_ahead_h: t_ahead_h
    :return: -
    """
    kpi_metrics = list(kpis[t_ahead_h].keys())
    # Set to nan
    for kpi in kpi_metrics:
        if kpi != "completeness":
            kpis[t_ahead_h].update({kpi: np.nan})



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import pandas as pd
import plotly
from ktpbase.database import DataBase
from ktpbase.log import logging

from openstf.feature_engineering.capacity_prognoses_features import (
    apply_capacity_features,
)
from openstf.model.capacity.model import CapacityPredictionModel
from openstf.model.capacity.utils import prepare_prediction_data, visualize_predictions
from openstf.model.serializer.xgboost.xgboost import XGBModelSerializer


def predict_capacity_prognosis(pj, datetime_start, datetime_end, y_hor=None):
    """Predict capacity prognoses for specific prediction job

    Args:
        pj: (dict) prediction job
        datetime_start: (datetime.datetime) start date
        datetime_end: (datetime.datetime) end date
        y_hor: (list): prediction horizons
    """
    if y_hor is None:
        y_hor = list(range(13))
    db = DataBase()

    # initialize logging
    logger = logging.get_logger(__name__)

    # get input data
    logger.info("Get input data")
    load_data = db.get_load_pid(
        pid=pj["id"], datetime_start=datetime_start, datetime_end=datetime_end
    )
    load_profiles_data = db.get_tdcv_load_profiles(datetime_start, datetime_end)
    load_profile_names = list(load_profiles_data.columns)
    input_data = pd.concat([load_data, load_profiles_data], axis=1)

    # apply features
    logger.info("Apply capacity features")
    feature_data, _ = apply_capacity_features(
        input_data, y_col="load_max", y_hor=y_hor, load_profile_names=load_profile_names
    )

    # prepare prediction points
    logger.info("Prepare prediction data")
    x = prepare_prediction_data(feature_data, y_col="load_max", y_hor=y_hor)

    # create model
    logger.info("Create capacity prognosis model")
    model = CapacityPredictionModel()

    # load model
    serializer = XGBModelSerializer()
    directory = serializer.trained_models_folder / str(pj["id"]) / "capacity"
    logger.info("Loading model", model_directory=directory)
    model.load(directory=directory)

    # predict
    logger.info("Make prediction")
    y_pred, y_pred_prob = model.predict(x)

    # visualize predictions
    logger.info("Visualize prediction")
    fig = visualize_predictions(df=y_pred_prob, classes=model.classes)

    # save figure
    plotly.offline.plot(
        fig,
        filename=str(directory / "capacity_prognoses.html"),
        show_link=False,
        auto_open=False,
    )
    logger.info("Sucessfully made capacity prediction")



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from abc import ABC, abstractmethod

import numpy as np
import pandas as pd

from openstf.model.general import MLModelType
from openstf.model.serializer.creator import ModelSerializerCreator


class AbstractModelTrainer(ABC):
    def __init__(self, pj):
        super().__init__()
        self.pj = pj
        self.trained_model = None
        self.old_model = None
        self.confidence_interval = None
        self.figs = None
        self.hyper_parameters = None

        self.serializer = ModelSerializerCreator().create_model_serializer(
            MLModelType(self.pj["model"])
        )
        self.old_model, old_model_location = self.serializer.load(self.pj["id"])
        self.old_model_age = self.serializer.determine_model_age_from_path(
            old_model_location
        )

    @property
    @abstractmethod
    def feature_importance(self):
        """Abstract method that gets feature importances of trained model.

        Returns:
            pandas.DataFrame: A DataFrame describing the feature importances of
            the trained model. At least the following columns should be
            included::

            index (str): Label of feature
            importance (float): Indication of how important the feature was to
            the trained model. Use only non-negative values. Does not need to
            sum to 1.
            weight (float): Indication of how often the feature was used in the
            trained model. Use only non-negative values. Should be normalized to
            sum to 1.

            If the model only supports a single importance/weight indication,
            i.e linear regression, set both columns to this value.

        """
        print(
            "This is an abstract class please use a derivative class where a \
            get_feature_importance method is implemented."
        )

    @abstractmethod
    def train(self, train_data, validation_data, callbacks=None):
        """Abstract method to train a model based on train and validation data.

        The method should add the newly trained model to the trained_model
        attribute.

        Args:
            train_data (pandas.DataFrame): The train data
            validation_data (pandas.DataFrame): The validation data
            callbacks (list of callable): List of callback functions that can be
                called at the end of each training iteration

        """
        print(
            "This is an abstract class please use a derivative class where a \
            train method is implemented."
        )

    @abstractmethod
    def better_than_old_model(self, test_data):
        """Abstract method that checks if new model is better than the old model.

        Args:
            test_data (pandas.DataFrame): The test data

        Returns:
            Bool: True if new model is better than old model, false otherwise

        """
        print(
            "This is an abstract class please use a derivative class where a \
            better_than_old_model method is implemented."
        )
        return

    @abstractmethod
    def store_model(self):
        """Abstract method that should store the model.

        The general implementation is to just call _store_model. However this
        function exists if some other steps must be done before _store_model can
        be called.

        """
        print(
            "This is an abstract class please use a derivative class where a \
            store_model method is implemented."
        )

    @abstractmethod
    def hyper_params_objective(
        self,
        trial,
        error_function=None,
        clean_data_with_features=None,
        featuresets=None,
        optimize_peak_prediction=False,
    ):
        print(
            "This is an abstract class please use a derivative class where a \
            get_corrections method is implemented."
        )

    def _store_model(self):
        """
            Method that stores a newly trained model to persistent storage.

        Args:
            model_type: ModelType enum that specifies to look for a component model
            name: str() name of the component
        """
        self.serializer.save(
            self.pj["id"], self.trained_model, corrections=self.confidence_interval
        )

    @staticmethod
    def _calculate_confidence_interval(realised, predicted):
        """Protected static method to calculate the corrections for a model

        Args:
            realised: pd.series with realised load
            predicted: pd.series with load predicted by new model

        Returns:
            pd.DataFrame: with model corrections
        """
        result = pd.DataFrame(index=range(24), columns=["stdev", "hour"])
        # Calculate the error for each predicted point
        error = realised - predicted
        error.index = error.index.hour  # Hour only, remove the rest
        # For the time starts with 00, 01, 02, etc. TODO (MAKE MORE ELEGANT SOLUTION THAN A LOOP)
        for hour in range(24):
            hour_error = error[error.index == hour]
            result["stdev"].iloc[hour] = np.std(hour_error)
            result["hour"].iloc[hour] = hour

        result = result.astype("float")

        return result



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pandas as pd
from ktpbase.log import logging

from openstf.feature_engineering import apply_features
from openstf.model.general import ForecastType


def normalize_and_convert_weather_data_for_splitting(weather_data):
    """Normalize and converts weather data for use in energy splitting.

    Args:
        weather_data (pd.DataFrame): Weather data with "windspeed_100m" and "radiation".

    Returns:
         pd.DataFrame: Dataframe with "windpower" and "radiation" columns.

    """

    # Check we have "windspeed_100m" and "radiation" available
    if not all(
        elem in weather_data.columns for elem in ["windspeed_100m", "radiation"]
    ):
        raise ValueError("weather data does not contain required data!")

    # Prepare output dataframe
    output_dataframe = pd.DataFrame()

    # Normalize weather data
    output_dataframe["radiation"] = (
        weather_data["radiation"]
        / np.percentile(weather_data["radiation"].dropna(), 99.0)
        * -1
    )
    wind_ref = apply_features.calculate_windspeed_at_hubheight(
        weather_data["windspeed_100m"], fromheight=100
    )
    wind_ref = wind_ref / np.abs(np.amax(wind_ref)) * -1

    output_dataframe["windpower"] = wind_ref

    return output_dataframe


def split_forecast_in_components(forecast, weather_data, split_coefs):
    """Function that makes estimates of energy components based on given forecast,
        previously determine splitting coefficients and relevant weather data.

    Args:
        forecast(pd.DataFrame): KTP load forecast
        weather_data (pd.DataFrame): Weather data for energy splitting, at least:
            "windspeed_100m" and "radiation"
        split_coefs (dict): Previously determined splitting coefs for prediction job

    Returns:
        dict: Forecast dataframe for each component

    """

    # Normalize weather data
    weather_ref_profiles = normalize_and_convert_weather_data_for_splitting(
        weather_data
    )

    # Check input
    if not all(
        elem in ["windpower", "radiation"]
        for elem in list(weather_ref_profiles.columns)
    ):
        raise ValueError("weather data does not contain required data!")

    # Merge to ensure datetime index is the same
    weather_ref_profiles = forecast.merge(
        weather_ref_profiles, how="outer", right_index=True, left_index=True
    )
    # Drop rows with duplicate indices
    weather_ref_profiles = weather_ref_profiles[
        ~weather_ref_profiles.index.duplicated()
    ]
    weather_ref_profiles.replace([np.inf, -np.inf], np.nan).dropna(inplace=True)

    # Prepare output dictionary and list of forecast types
    components = forecast.copy(deep=True)

    # Calculate profiles of estimated components
    components["forecast_wind_on_shore"] = (
        split_coefs["wind_ref"] * weather_ref_profiles["windpower"]
    )
    components["forecast_solar"] = (
        split_coefs["pv_ref"] * weather_ref_profiles["radiation"]
    )
    components["forecast_other"] = (
        weather_ref_profiles["forecast"]
        - components["forecast_solar"]
        - components["forecast_wind_on_shore"]
    )

    # Check that sign of production components is negative and not positive, change if sign is wrong
    if components["forecast_wind_on_shore"].sum() > 0:
        raise Exception("Sign of estimated wind_on_shore component is positive!")
    if components["forecast_solar"].sum() > 0:
        raise Exception("Sign of estimated solar component is positive!")

    # Post process predictions to ensure realistic values
    components["forecast_solar"] = post_process_wind_solar(
        components["forecast_solar"], ForecastType.SOLAR
    )
    components["forecast_wind_on_shore"] = post_process_wind_solar(
        components["forecast_wind_on_shore"], ForecastType.WIND
    )

    return components.drop("forecast", axis=1).dropna()


def post_process_wind_solar(forecast: pd.Series, forecast_type):
    """Function that caries out postprocessing for wind and solar power generators.

        As these points will always produce energy, predicted energy consumption is
        set to zero. This function will automatically detect the sign as this can
        vary from case to case. The largest volume either positive or negative is
        assumed to be the energy production and the other side the energy consumption.

    Args:
        forecast (pd.Series): Series with forecast data.
        forecast_type (ForecastType): Specifies the type of forecast. This can be retrieved
            from the prediction job as pj['typ']

    Returns:
        forecast (pd.DataFrame): post-processed forecast.

    """
    logger = logging.get_logger(__name__)

    if forecast_type not in [ForecastType.WIND, ForecastType.SOLAR]:
        return forecast

    forecast_data_sum = forecast.sum()
    # Determine sign of sum
    if forecast_data_sum > 0:
        # Set all values smaller than zero to zero, since this is not realistic
        forecast.loc[forecast < 0] = 0
    elif forecast_data_sum < 0:
        # Likewise for all values greater than zero
        forecast.loc[forecast > 0] = 0
    else:
        logger.warning(
            f"Could not determine sign of the forecast, skip post-processing. Sum was {forecast_data_sum}"
        )

    # write changed back to forecast
    return forecast



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import json
import os
from pathlib import Path

from ktpbase.config.config import ConfigManager


def get_module_name(module_path):
    # /home/user/git/project/package/module.py -> module
    module_name = ".".join(str(module_path).split(os.path.sep)[-1].split(".")[:-1])
    return module_name


def load_status_file():
    """
    This function loads the json status file
    :return: dictionary (status object)
    """
    status_path = Path(ConfigManager.get_instance().paths.webroot)
    status_file = "status.json"
    # Check if current status exists
    if os.path.isfile(status_path / status_file):
        # Load current status
        status_file = open(status_path / status_file, "r")
        status_file_string = status_file.read()
        try:
            status_dict = json.loads(status_file_string)
        except json.JSONDecodeError:
            status_file_string = status_file_string.replace("'", '"')
            status_dict = json.loads(status_file_string)
    else:
        # Create empty status
        status_dict = {}
    return status_dict


def write_status_file(status):
    """
    This function writes the given status object to a json file
    :param status: dictionary
    :return: None
    """
    status_path = Path(ConfigManager.get_instance().paths.webroot)
    status_file = "status.json"

    # If path does not exists, create it
    if not os.path.exists(status_path):
        os.makedirs(status_path)

    # Write status file
    with open(status_path / status_file, "w") as json_file:
        json.dump(status, json_file)


def check_status_change(name, new_code):
    """
    Check if the old status code equals the new code.
    If no status if found it is assumed that the old code is 0
    :param name: name of the status field
    :param new_code: int
    :return: True if changed
    """

    # Load previous status
    status = load_status_file()

    # Get old status
    old_code = 0
    if name in status:
        old_code = status[name]

    # Equality check
    return old_code != new_code


def update_status_change(name, new_code):
    """

    :param name: name of the status field
    :param new_code: int
    :return: None
    """

    # Load status
    status = load_status_file()

    # Update status
    status[name] = new_code

    # Write status
    write_status_file(status)


def convert_string_args_to_dict_args(argsstring):
    """Helper function that converts a string of arguments to a dict of arguments.
    Coverts inputs to floats if possible. Assumes args are seperated by ',' or ', '.

    input: str

    output: dict(args=[arg1, ..., argN], kwargs=dict(kwarg1=val1, ..., kwargN=valN)

    example:
        '5, seconds' -> dict(args=[5, 'seconds'])
        '3, val=blub -> dict(args=[3], kwargs=dict(val='blub')"""

    # convert ', 'to ',' and split args on ','
    args = argsstring.replace(", ", ",").split(",")
    # Make list of all items without a '=' -> the args
    arg_list = [x for x in args if "=" not in x]

    # Make a list of all items with a '=' -> the kwargs
    # kwarglist should be like ['key1=val1','key2 = val2']
    kwarg_list = [x for x in args if x not in arg_list]
    # Now let's turn the list into a dict. e.g. {val1:'key1', val2:'key2'}
    kwargs_dict = {
        key: value
        for key, value in [
            kwarg_pair.replace(" = ", "=").split("=") for kwarg_pair in kwarg_list
        ]
    }

    # Lastly, convert values to floats if possible
    arg_list = [float(x) if x.replace(".", "").isdigit() else x for x in arg_list]
    kwargs_dict = {
        key: float(value) if value.replace(".", "").isdigit() else value
        for key, value in kwargs_dict.items()
    }

    return dict(args=arg_list, kwargs=kwargs_dict)


def interpret_string_as_functions(function, args, kwargs):
    """Interpret a given function with args and kwargs to build a string suitable for eval
    input:
        function = str
        args = list(arg1, ..., argN)
        kwargs = dict(kwarg1=val1, ..., kwargN=valN)

    return:
        function(arg1, ..., argN, kwarg1=val1, ..., kwargN=valN)
    """
    arg_string = ", ".join(
        [str(x) if type(x) in [float, int] else "'{}'".format(x) for x in args]
    )
    if arg_string == "":
        arg_string = arg_string + "db"
    else:
        arg_string = arg_string + ", db"

    kwarg_string = ""
    # build sorted list of kwarg keys; otherwise not able to unittest
    kwargkeys = sorted(kwargs.keys())
    for key in kwargkeys:
        value = kwargs[key]
        # add quotes if value is a string
        if not isinstance(value, (int, float)):
            value = "'" + value + "'"
        kwarg_string += "{}={}, ".format(key, value)
    # remove last ,
    if len(kwarg_string) > 2:
        kwarg_string = kwarg_string[:-2]

    # add comma between argstring and kwargstring if argstring not empty and kwargstring not empty
    if (len(arg_string) > 0) and (len(kwarg_string) > 0):
        arg_string += ", "

    # Build result
    result = function + "(" + arg_string + kwarg_string + ")"
    return result



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import timedelta

import numpy as np
from ktpbase.log import logging

from openstf.feature_engineering import general


def validate(data):
    logger = logging.get_logger(__name__)
    # Drop 'false' measurements. e.g. where load appears to be constant.
    threshold = 6 * 4  # number of repeated values
    data = general.nan_repeated(data, max_length=threshold, column_name=data.columns[0])
    num_const_load_values = len(data) - len(data.iloc[:, 0].dropna())
    logger.debug(
        f"Changed {num_const_load_values} values of constant load to NA.",
        num_const_load_values=num_const_load_values,
    )
    return data


def clean(data):
    logger = logging.get_logger(__name__)
    data = data[data.index.min() + timedelta(weeks=2) :]
    len_original = len(data)
    # TODO Look into this
    # Remove where load is NA # # df.dropna?
    data = data.loc[np.isnan(data.iloc[:, 0]) != True, :]  # noqa E712
    num_removed_values = len_original - len(data)
    logger.debug(
        f"Removed {num_removed_values} NA values", num_removed_values=num_removed_values
    )
    return data



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# -*- coding: utf-8 -*-
""" This module contains all general feature engineering functions"""
import re
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import scipy

from openstf.feature_engineering import apply_features


def extract_minute_features(feature_names):
    """Creates a list of features that were used during training of the input model
    Args:
        feature_names (list[str]): Feature names

    Returns:
        minute_list (list[int]): list of minute lags that were used as features during training
    """

    minutes_list = []
    for feature in feature_names:
        # TODO Should this module really know how the feature names are written?
        # Perhaps better if this function uses a list with all possible minute features
        # or a tempalte. In any case the knowlegde should come from the module which
        # creates the features.
        m = re.search(r"T-(\d+)min", feature)

        if m is None:
            continue
        else:
            minutes_list.append(int(m[1]))

    return minutes_list


def calc_completeness(df, weights=None, time_delayed=False, homogenise=True):
    """Calculate the (weighted) completeness of a dataframe.

    NOTE: NA values count as incomplete

    Args:
        df (pd.DataFrame): Dataframe with a datetimeIndex index
        weights: Array-compatible with size equal to columns of df.
            used to weight the completeness of each column
        time_delayed (bool): Should there be a correction for T-x columns
        homogenise (bool): Should the index be resampled to median time delta -
            only available for DatetimeIndex

    Returns:
        float: Completeness
    """

    if weights is None:
        weights = np.array([1] * len(df.columns))
    weights = np.array(weights)

    if homogenise and isinstance(df.index, pd.DatetimeIndex) and len(df) > 0:

        median_timediff = int(
            df.reset_index().iloc[:, 0].diff().median().total_seconds() / 60.0
        )
        df = df.resample("{:d}T".format(median_timediff)).mean()

    if time_delayed is False:
        # Calculate completeness
        # Completeness per column
        completeness_per_column = df.count() / len(df)

    # if timeDelayed is True, we correct that time-delayed columns
    # also in the best case will have NA values. E.g. T-2d is not available
    # for times ahead of more than 2 days
    elif time_delayed:
        # assume 15 minute forecast resolution
        # timecols: {delay:number of points expected to be missing}
        # number of points expected to be missing = numberOfPointsUpToTwoDaysAhead - numberOfPointsAvailable
        timecols = {
            x: len(df) - eval(x[2:].replace("min", "/60").replace("d", "*24.0")) / 0.25
            for x in df.columns
            if x[:2] == "T-"
        }

        non_na_count = df.count()
        for col, value in timecols.items():
            non_na_count[col] += value

        # Correct for APX being only expected to be available up to 24h
        if "APX" in non_na_count.index:
            non_na_count["APX"] += max([len(df) - 96, 0])

        completeness_per_column = non_na_count / len(df)

    # scale to weights and normalize
    completeness = (completeness_per_column * weights).sum() / weights.sum()

    return completeness


def nan_repeated(df, max_length, column_name):
    """
    This function replaces repeating values (sequentially repeating values),
    which repeat longer than a set max_length (in data points) with NaNs.

    Args:
        df (pandas.DataFrame): Data from which you would like to set repeating values to nan
        max_length (int): If a value repeats more often, sequentially, than this value, all those points are set to NaN
        column_name (string): the pandas dataframe column name of the column you want to process

    Rrturns:
        pandas.DataFrame: data, similar to df, with the desired values set to NaN.
    """
    data = df.copy(deep=True)
    indices = []
    old_value = -1000000000000
    value = 0
    for index, r in data.iterrows():
        value = r[column_name]
        if value == old_value:
            indices.append(index)
        elif (value != old_value) & (len(indices) > max_length):
            indices = indices[max_length:]
            data.at[indices, column_name] = np.nan
            indices = []
            indices.append(index)
        elif (value != old_value) & (len(indices) <= max_length):
            indices = []
            indices.append(index)
        old_value = value
    if len(indices) > max_length:
        data.at[indices, column_name] = np.nan
    return data


def get_preprocessed_data(
    pj,
    db,
    datetime_start=datetime.utcnow() - timedelta(days=90),
    datetime_end=datetime.utcnow(),
):
    """Retrieves input data frame and pre-process it (calculate features, correct constant load, drop NAs etc.)
    Input:
        pj, the current prediction job
    Output:
        table: pd.DataFrame
        completeness: percentage of data kept after pre-processing"""
    location = [pj["lat"], pj["lon"]]

    if "+" in pj["description"]:
        name_var = pj["description"]
    else:
        name_var = pj["sid"]

    table = db.get_model_input(
        pid=pj["id"],
        location=location,
        datetime_start=datetime_start,
        datetime_end=datetime_end,
    )

    if "load" not in table.columns:
        raise SystemError("No Load returned for {}".format(pj["description"]))

    # !!!!!
    # Work-around for cityzen data
    if name_var.split("_")[0] == "cable":
        print("NOTE! Rescaling input data!")
        table.iloc[:, 0] *= 1000.0

    # Drop 'false' measurements. e.g. where load appears to be constant.
    threshold = 6 * 4  # number of repeated values
    table = nan_repeated(table, threshold, table.columns[0])
    const_load_values = len(table) - len(table.iloc[:, 0].dropna())
    print("Changed {} values of constant load to NA.".format(const_load_values))

    # Apply model features to get the full table
    horizons = [0.25, 47]
    table = apply_features.apply_multiple_horizon_features(table, h_aheads=horizons)
    # Drop first rows where not enough data was present to make T-14d for example
    # For now, use fixed limit of first two weeks

    completeness = calc_completeness(table, time_delayed=True)

    return table, completeness


def calc_norm(data, how="max", add_to_df=True):
    """This script calculates the norm of a given dataset.
    Input:
        - data: pd.DataFrame(index = datetime, columns = [load])
        - how: str can be any function from numpy, recognized by np.'how'
        Optional:
        - add_to_df: Bool, add the norm to the data

    Output:
        - pd.DataFrame(index = datetime, columns = [load])
    NB: range of datetime of input is equal to range of datetime of output

    Example
    import pandas as pd
    import numpy as np
    index = pd.date_range(start = "2017-01-01 09:00:00", freq = '15T', periods = 200)
    data = pd.DataFrame(index = index,
                        data = dict(load=np.sin(index.hour/24*np.pi)*np.random.uniform(0.7,1.7, 200)))"""

    colname = list(data)[0]
    if how == "max":
        df = data.groupby(data.index.time).apply(lambda x: x.max(skipna=True))
    if how == "mean":
        df = data.groupby(data.index.time).apply(lambda x: x.mean(skipna=True))

    # rename
    df.rename(columns={colname: "Norm"}, inplace=True)

    # Merge to dataframe if add_to_df == True
    if add_to_df:
        df = data.merge(df, left_on=data.index.time, right_index=True)[
            [colname, "Norm"]
        ].sort_index()

    return df


def apply_persistence(data, how="mean", smooth_entries=4, add_to_df=True, colname=None):
    """This script calculates the persistence forecast
    Input:
        - data: pd.DataFrame(index = datetime, columns = [load]), datetime is expected to have historic values, as well as NA values
        Optional:
        - how: str, how to determine the norm (abs or mean)
        - smoothEntries: int, number of historic entries over which the persistence is smoothed
        - add_to_df: Bool, add the forecast to the data
        - option of specifying colname if load is not first column

    Output:
        - pd.DataFrame(index = datetime, columns = [(load,) persistence])
    NB: range of datetime of input is equal to range of datetime of output

    Example
    import pandas as pd
    import numpy as np
    index = pd.date_range(start = "2017-01-01 09:00:00", freq = '15T', periods = 300)
    data = pd.DataFrame(index = index,
                        data = dict(load=np.sin(index.hour/24*np.pi)*np.random.uniform(0.7,1.7, 300)))
    data.loc[200:,"load"] = np.nan"""

    data = data.sort_index()

    if colname is None:
        colname = list(data)[0]

    df = calc_norm(data, how=how, add_to_df=True)

    # this selects the last non NA values
    last_entries = df.loc[df[colname].notnull()][-smooth_entries:]

    norm_mean = last_entries.Norm.mean()
    if norm_mean == 0:
        norm_mean = 1

    factor = last_entries[colname].mean() / norm_mean
    df["persistence"] = df.Norm * factor

    if add_to_df:
        df = df[[colname, "persistence"]]
    else:
        df = df[["persistence"]]

    return df


def apply_fit_insol(data, add_to_df=True, hours_delta=None, polynomial=False):
    """This model fits insolation to PV yield and uses this fit to forecast PV yield.
    It uses a 2nd order polynomial

    Input:
        - data: pd.DataFrame(index = datetime, columns = [load, insolation])
        Optional:
        - hoursDelta: period of forecast in hours [int] (e.g. every 6 hours for KNMI)
        - addToDF: Bool, add the norm to the data

    Output:
        - pd.DataFrame(index = datetime, columns = [(load), forecaopenstfitInsol])
    NB: range of datetime of input is equal to range of datetime of output

    Example
    import pandas as pd
    import numpy as np
    index = pd.date_range(start = "2017-01-01 09:00:00", freq = '15T', periods = 300)
    data = pd.DataFrame(index = index,
                        data = dict(load=np.sin(index.hour/24*np.pi)*np.random.uniform(0.7,1.7, len(index))))
    data['insolation'] = data.load * np.random.uniform(0.8, 1.2, len(index)) + 0.1
    data.loc[int(len(index)/3*2):,"load"] = np.NaN"""

    colname = list(data)[0]

    # Define subset, only keep non-NaN values and the most recent forecasts
    # This ensures a good training set
    if hours_delta is None:
        subset = data.loc[(data[colname].notnull()) & (data[colname] > 0)]
    else:
        subset = data.loc[
            (data[colname].notnull())
            & (data[colname] > 0)
            & (data["tAhead"] < timedelta(hours=hours_delta))
            & (data["tAhead"] >= timedelta(hours=0))
        ]

    def linear_fun(coefs, values):
        return coefs[0] * values + coefs[1]

    def second_order_poly(coefs, values):
        return coefs[0] * values ** 2 + coefs[1] * values + coefs[2]

    # Define function to be minimized and subsequently minimize this function
    if polynomial:
        # Define starting guess
        x0 = [1, 1, 0]  # ax**2 + bx + c.
        fun = (
            lambda x: (second_order_poly(x, subset.insolation) - subset[colname])
            .abs()
            .mean()
        )
        # , bounds = bnds, constraints = cons)
        res = scipy.optimize.minimize(fun, x0)
        # Apply fit
        df = second_order_poly(res.x, data[["insolation"]]).rename(
            columns=dict(insolation="forecaopenstfitInsol")
        )

    else:
        x0 = [1, 0]
        fun = (
            lambda x: (linear_fun(x, subset.insolation) - subset[colname]).abs().mean()
        )
        res = scipy.optimize.minimize(fun, x0)
        df = linear_fun(res.x, data[["insolation"]]).rename(
            columns=dict(insolation="forecaopenstfitInsol")
        )

    # Merge to dataframe if addToDF == True
    if add_to_df:
        if hours_delta is None:
            df = data.merge(df, left_index=True, right_index=True)
        else:
            df = pd.concat([data, df], axis=1)

    return df


def add_missing_feature_columns(input_data, featurelist):
    """Adds feature column for features in the featurelist.

    Add feature columns for features in the feature list if these columns don't
    exist in the input data. If a column is added, its value is set to NaN.
    """
    missing_features = [f for f in featurelist if f not in list(input_data)]

    for feature in missing_features:
        print(f"Warning: adding NaN column for missing feature: {feature}")
        input_data[feature] = np.nan

    return input_data


def remove_features_not_in_set(data_with_features, featureset):
    """Remove all features not in the given featureset.

    Args:
        data_with_features (pandas.DataFrame): Data with features
        featureset_name (list): Name of the featureset to keep (if exists)

    Returns:
        pandas.DataFrame: Data with features not in featureset removed
    """
    # always keep the load and the Horizon column
    columns_to_keep = ["load", "Horizon"]

    features_to_keep = (
        columns_to_keep
        +
        # this makes sure we only keep features that actually exist
        [f for f in featureset if f in data_with_features.columns]
    )

    return data_with_features[features_to_keep]



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timedelta
from pathlib import Path

from ktpbase.database import DataBase
from ktpbase.config.config import ConfigManager


def make_report_pj(pj):
    """Makes a csv report of realized load and forecasted load for a prediction job.
    The csv is saved in the icarus visuals folder.

    Args:
        pj (dict): Prediction job
    """
    # Intitialize databse
    db = DataBase()
    config = ConfigManager.get_instance()

    output_dir = Path(config.paths.webroot) / "ato_reports"
    output_file = "report_{}.csv".format(pj["name"])

    # Define start and end of realized date
    start = (datetime.utcnow() - timedelta(days=7)).replace(
        tzinfo=None, minute=0, hour=0
    )

    end_realized = datetime.utcnow()

    # Define end of forecast period
    end_forecast = (end_realized + timedelta(days=2)).replace(minute=0, hour=0)

    # Get realized load
    load = db.get_load_pid(pj["id"], start, end_realized)
    # Get forecast
    forecast = db.get_predicted_load_tahead(pj, start, end_forecast, t_ahead="24H")

    # Merge forecast and load into one dataframe
    forecast = forecast[["forecast_24.0h"]]
    forecast = forecast.rename(columns={"forecast_24.0h": "day_ahead_forecast"})
    forecast = forecast.dropna()
    load = load.merge(forecast, how="outer", left_index=True, right_index=True)

    # Convert timezone from utc to CET as end users expect this time zone
    load.index = load.index.tz_convert("CET")
    load.index.name = "datetime"
    # Rename to make extra clear this is realised
    load = load.rename(columns={"load": "realised_load"})

    output_dir.mkdir(parents=True, exist_ok=True)

    # Write to csv
    load.to_csv(output_dir / output_file)  # can join path elements with / operator

    return load



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from abc import ABC, abstractmethod
from datetime import datetime, timedelta, timezone

import numpy as np
import pandas as pd

from openstf.model.general import ForecastType, MLModelType
from openstf.model.serializer.creator import ModelSerializerCreator

FLATLINER_THRESHOLD = 6


class AbstractPredictionModel(ABC):
    def __init__(
        self, pj, forecast_type, trained_model=None, trained_confidence_df=None
    ):
        super().__init__()
        self.pj = pj
        self.model_type = MLModelType(pj["model"])
        self.forecast_type = forecast_type
        self.serializer = ModelSerializerCreator.create_model_serializer(
            model_type=self.model_type
        )
        self.confidence_df = trained_confidence_df
        self.model = trained_model
        self.model_file_path = None

        if self._bool_load_model_from_disk():
            self.model, self.model_file_path = self.serializer.load(pid=self.pj["id"])

    def _bool_load_model_from_disk(self):
        """Determine if model should be loaded from disk"""
        return self.forecast_type is not ForecastType.BASECASE and self.model is None

    @property
    @abstractmethod
    def feature_names(self):
        pass

    @abstractmethod
    def calculate_completeness(self, forcast_input_data):
        pass

    @abstractmethod
    def make_forecast(self, forcast_input_data):
        pass

    def make_fallback_forecast(self, forecast_input_data, load_data):
        if self.forecast_type is ForecastType.BASECASE:
            raise ValueError(
                "Can't make a fallback prediction for forecast type basecase"
            )

        self.logger.warning(
            "Using failback forecast (*high forecast*)", forecast_type="fallback"
        )
        forecast = self.predict_fallback(
            forecast_index=forecast_input_data.index, load=load_data
        )

        # Add corrections (if file exists and is not empty)
        forecast = self._add_standard_deviation_to_forecast(forecast)
        # Add prediciont job propertie to forecast
        forecast = self.add_prediction_job_properties_to_forecast(
            pj=self.pj,
            forecast=forecast,
            algorithm_type=str(self.model_file_path),
            forecast_type=self.forecast_type,
        )
        return forecast

    @staticmethod
    def make_basecase_forecast(pj, historic_load, overwrite_delay_hours=48):
        """Make a 'basecase' forecast

            Result is writen to database for all forecasts further in time than
            overwrite_delay_hours. Idea is that if all else fails, this forecasts is
            still available.

            Basecase example: the load of last week.

            As 'quality', the value 'not_renewed' is used.

        Args:
            historic_load (pd.DataFrame): Historic load
            overwrite_delay_hours (float): times before this in the future are not
                forecasted

        Returns:
            pd.DataFrame: Basecase forecast (which was written to the database)
        """
        # - Make forecast
        # Make basecase forecast: Use load of last week
        basecase_forecast = historic_load.shift(7, "D")

        # Maybe there is still missing data, for example if the cdb has been down for a
        # while in this case, use the load of 2 weeks before
        basecase_forecast = basecase_forecast.append(historic_load.shift(14, "D"))
        basecase_forecast = basecase_forecast[
            np.invert(basecase_forecast.index.duplicated())
        ]

        # - Post Process
        # Don't update first 48 hours
        forecast_start = datetime.now(timezone.utc) + timedelta(
            hours=overwrite_delay_hours
        )
        basecase_forecast = basecase_forecast[forecast_start:]
        # Don't update nan values
        basecase_forecast.dropna(inplace=True)
        # rename
        basecase_forecast = basecase_forecast.rename(columns=dict(load="forecast"))

        # Also make a basecase forecast for the forecast_other component. This will make a
        # simple basecase components forecast available and ensures that the sum of
        # the components (other, wind and solar) is equal to the normal basecase forecast
        basecase_forecast["forecast_other"] = basecase_forecast["forecast"]

        # Estimate the stdev a bit smart
        # use the stdev of the hour for historic_load
        std_per_hour = (
            historic_load.groupby(historic_load.index.hour)
            .std()
            .rename(columns=dict(load="stdev"))
        )
        basecase_forecast["hour"] = basecase_forecast.index.hour
        basecase_forecast = basecase_forecast.merge(
            std_per_hour, left_on="hour", right_index=True
        )
        del basecase_forecast["hour"]

        # - Add properties to forecast
        basecase_forecast = (
            AbstractPredictionModel.add_prediction_job_properties_to_forecast(
                pj=pj,
                forecast=basecase_forecast,
                algorithm_type="basecase_lastweek",
                forecast_quality="not_renewed",
            )
        )

        return basecase_forecast.sort_index()

    @staticmethod
    def predict_fallback(forecast_index, load):
        """Make a fall back forecast
            Use historic profile of most extreme day.

            Set the value of the forecast 'quality' column to 'substituted'

        Args:
            forecast_index (pandas.DatetimeIndex): Index required for the forecast
            load (pandas.DataFrame): load

        Raises:
            RuntimeError: When the most important feature does not start with
                T-, wind or radi

        Returns:
            pandas.DataFrame: Fallback forecast DataFrame with columns:
                'forecast', 'quality'
        """
        # Check if load is completely empty
        if len(load.dropna()) == 0:
            raise Exception("No historic load data available")

        # Find most extreme historic day
        rel_date = load.load.idxmax().date()
        dayprof = load[str(rel_date)].copy()
        dayprof["time"] = dayprof.index.time

        forecast = pd.DataFrame(index=forecast_index)
        forecast["time"] = forecast.index.time
        forecast = (
            forecast.reset_index()
            .merge(dayprof, left_on="time", right_on="time", how="outer")
            .set_index("index")
        )
        forecast = forecast[["load"]].rename(columns=dict(load="forecast"))

        # Add a column quality.
        forecast["quality"] = "substituted"

        return forecast

    def _get_standard_deviation(self):

        # 1. used supplied (constructor) standard deviation if not None
        if self.confidence_df is not None:
            self.logger.info("Using supplied confidence_df as standard deviation")
            return self.confidence_df

        # 2. load standard deviation from disk (stored after training model)
        self.logger.info("Loading corrections from disk")
        corrections_file_path = (
            self.model_file_path.parent / f"{self.serializer.CORRECTIONS_FILENAME}"
        )
        try:
            corrections = pd.read_csv(corrections_file_path, index_col=0)
        except FileNotFoundError as e:
            self.logger.warning(
                "Did not add corrections, file not found. Try retraining the model "
                "with cor = True",
                exc_info=e,
            )
            return None

        if len(corrections) != len(corrections.dropna()):
            self.logger.warning("Correction file contains no values!")
            return None

        return corrections

    def _add_standard_deviation_to_forecast(self, forecast):
        """Add a standard deviation to a live forecast.

        The stdev for intermediate forecast horizons is interpolated.

        For the input standard_deviation, it is preferred that the forecast horizon is
        expressed in Hours, instead of 'Near/Far'. For now, Near/Far is supported,
        but this will be deprecated.

        Args:
            forecast (pd.DataFrame): Forecast DataFram with columns: "forecast"
            standard_deviation (pd.DataFrame): Standard deviation. DataFrame with columns:
                "hour", "horizon", "stdev"
            (optional) interpolate (str): Interpolation method, options: "exponential" or "linear"

        Returns:
            pd.DataFrame: Forecast with added standard deviation. DataFrame with columns:
                "forecast", "stdev"
        """

        if self.forecast_type is ForecastType.BASECASE:
            raise ValueError(
                "Can not make forecast corrections for forecast type basecase"
            )
        standard_deviation = self._get_standard_deviation()

        if standard_deviation is None:
            return forecast

        # -------- Moved from feature_engineering.add_stdev ------------------------- #
        # pivot
        stdev = standard_deviation.pivot_table(columns=["horizon"], index="hour")[
            "stdev"
        ]
        # Prepare input dataframes
        # Rename Near and Far to 0.25 and 47 respectively, if present.
        # Timehorizon in hours is preferred to less descriptive Near/Far
        if "Near" in stdev.columns:
            near = (forecast.index[1] - forecast.index[0]).total_seconds() / 3600.0
            # Try to infer for forecast df, else use a max of 48 hours
            far = min(
                48.0,
                (forecast.index.max() - forecast.index.min()).total_seconds() / 3600.0,
            )
            stdev.rename(columns={"Near": near, "Far": far}, inplace=True)
        else:
            near = stdev.columns.min()
            far = stdev.columns.max()

        forecast_copy = forecast.copy(deep=True)
        # add time ahead column if not already present
        if "tAhead" not in forecast_copy.columns:
            # Assume first datapoint is 'now'
            forecast_copy["tAhead"] = (
                forecast_copy.index - forecast_copy.index.min()
            ).total_seconds() / 3600.0
        # add helper column hour
        forecast_copy["hour"] = forecast_copy.index.hour

        # Define functions which can be used to approximate the error for in-between time horizons
        # Let's fit and exponential decay of accuracy
        def calc_exp_dec(t, stdev_row, near, far):
            # We use the formula sigma(t) = (1 - A * exp(-t/tau)) + b
            # Strictly speaking, tau is specific for each time series.
            # However, for simplicity, we use tau = Far/4.
            # This represents a situation where the stdev at 25% of the Far horizon,
            # has increased by two.
            tau = far / 4.0
            # Filling in the known sigma(Near) and sigma(Far) gives:
            sf, sn = stdev_row[far], stdev_row[near]
            A = (sf - sn) / ((1 - np.exp(-far / tau)) - (1 - np.exp(-near / tau)))
            b = sn - A * (1 - np.exp(-near / tau))
            return A * (1 - np.exp(-t / tau)) + b

        # Add stdev to forecast_copy dataframe
        forecast_copy["stdev"] = forecast_copy.apply(
            lambda x: calc_exp_dec(x.tAhead, stdev.loc[x.hour], near, far), axis=1
        )
        # -------- End moved from feature_engineering.add_stdev --------------------- #

        self.logger.info("Corrections added to forecast")

        return forecast_copy.drop(columns=["hour"])

    # NOTE idea is that this can perhaps be removed since its not required to make a
    # prediction but more specific to how we (Alliander/KTP) store things
    @staticmethod
    def add_prediction_job_properties_to_forecast(
        pj, forecast, algorithm_type, forecast_type=None, forecast_quality=None
    ):
        # self.logger.info("Postproces in preparation of storing")
        if forecast_type is None:
            forecast_type = pj["typ"]
        else:
            # get the value from the enum
            forecast_type = forecast_type.value

        # NOTE this field is only used when making the babasecase forecast and fallback
        if forecast_quality is not None:
            forecast["quality"] = forecast_quality

        # TODO rename prediction job typ to type
        # TODO algtype = model_file_path, perhaps we can find a more logical name
        # TODO perhaps better to make a forecast its own class!
        # TODO double check and sync this with make_basecase_forecast (other fields are added)
        # !!!!! TODO fix the requirement for customer
        forecast["pid"] = pj["id"]
        forecast["customer"] = pj["name"]
        forecast["description"] = pj["description"]
        forecast["type"] = forecast_type
        forecast["algtype"] = algorithm_type

        return forecast



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pandas as pd
import scipy
import xgboost as xgb
from ktpbase.log import logging

from openstf.model.prediction.xgboost.xgboost import XGBPredictionModel


class QuantileXGBPredictionModel(XGBPredictionModel):
    def __init__(
        self, pj, forecast_type, trained_model=None, trained_confidence_df=None
    ):
        super().__init__(pj, forecast_type, trained_model, trained_confidence_df)
        self.logger = logging.get_logger(self.__class__.__name__)

    @property
    def feature_names(self):
        return self.model.feature_names

    def make_forecast(self, forcast_input_data):
        """Make a forecast using a trained XGBoost model.

        Args:
            forcast_input_data (pandas.DataFrame): Forcast (section) input data

        Returns:
            pandas.DataFrame: The forecast DataFrame will have a 'forecast' column and
                a column for eacht percentile in the format:
            'quantile_Pxx' where xx represents the percentile. For example quantile_P50
                is the 50th percentile.

        """
        self.logger.info("Use XGBoost quantile model to make a forecast")

        forecast_request = xgb.DMatrix(forcast_input_data)

        model_forecast = self.model.predict(
            forecast_request, ntree_limit=self.model.best_ntree_limit
        )
        forecast = pd.DataFrame(
            index=forcast_input_data.index, data={"forecast": model_forecast}
        )

        # Loop over quantiles and add quantile forecasts
        for quantile in self.model.quantile_models.keys():
            quantile_key = f"quantile_P{quantile * 100:02.0f}"
            ntree_limit = self.model.quantile_models[quantile].best_ntree_limit
            quantile_forecast = self.model.predict(
                forecast_request, quantile=quantile, ntree_limit=ntree_limit
            )
            forecast[quantile_key] = quantile_forecast
            self.logger.info("Made quantile forecast", quantile=quantile)

        forecast = self._add_standard_deviation_to_forecast(forecast)

        # Add prediction job property to forecast
        forecast = self.add_prediction_job_properties_to_forecast(
            pj=self.pj,
            forecast=forecast,
            algorithm_type=str(self.model_file_path),
            forecast_type=self.forecast_type,
        )

        forecast_length = len(forecast)
        self.logger.info(
            f"Finished making quantile forecast with length '{forecast_length}'",
            forecast_length=forecast_length,
        )

        return forecast

    @staticmethod
    def _add_standard_deviation_to_forecast(forecast):
        """Calculate the mean stdev for a forecast dataframe with different quantiles present.

        Args:
            forecast (pd.DataFrame): DatetimeIndex, columns=['forecast', 'quantile_Pxx', ..., 'quantile_Pnn',]

        Returns:
            pd.DataFrame(DatetimeIndex, columns=['forecast', 'quantile_Pxx', ..., 'quantile_Pnn', 'stdev']

        Raises:
            ValueError: when forecast contains no quantile columns
        """
        standard_deviations = []
        num_quantile_columns = 0

        # calculate standard deviation for each quantile
        for column_name in filter(
            lambda c: c.startswith("quantile_P"), forecast.columns
        ):
            quantile_value = forecast[column_name]
            quantile = float(column_name.split("quantile_P")[1]) / 100.0
            if quantile == 0.5:  # For this value we get inf values
                continue
            stdev = (quantile_value - forecast["forecast"]) / scipy.stats.norm.ppf(
                quantile
            )

            standard_deviations.append(stdev)
            num_quantile_columns += 1

        if num_quantile_columns == 0:
            raise ValueError(
                "Can not calculate standard deviation, forecast does not contain quantile columns"
            )

        # calculate average standard deviation
        standard_deviation = np.nansum(standard_deviations, axis=0) / len(
            standard_deviations
        )
        forecast["stdev"] = standard_deviation

        return forecast

    def predict_fallback(self, forecast_index, load):
        """Overwrite default fallbackforecast to include quantiles.
        Use default fallbackforecast as startingpoint
        Quantiles are based on percentiles of hourly-grouped load_data
        Which quantiles are calculated are hard-coded for now
        TODO: make hardcoded quantiles not hardcoded

        Args:
            -forecast_index: pd.DatetimeIndex, used to determine new forecast timerange
            - load: pd.DataFrame(DatetimeIndex, columns=[load])

        Returns:
            pd.DataFrame(DatetimeIndex, forecast, quantile_Px,...)
        """

        # Use default
        forecast = super().predict_fallback(forecast_index, load)

        # Calculate quantiles per hour of day - list with items <1 can be copied from database
        quantiles = [x * 100 for x in [0.05, 0.1, 0.3, 0.5, 0.7, 0.9, 0.95]]
        q_vals = load.groupby(load.index.hour).apply(
            lambda x: np.percentile(x, quantiles)
        )

        q_vals = pd.DataFrame(
            q_vals.values.tolist(),
            index=q_vals.index,
            columns=[f"quantile_P{quantile:02.0f}" for quantile in quantiles],
        )

        # Add quantiles to the forecast
        forecast = forecast.merge(q_vals, left_on=forecast.index.hour, right_index=True)
        forecast = forecast.loc[:, [x for x in forecast.columns if x != "key_0"]]
        return forecast



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import os
from datetime import datetime, timedelta
from pathlib import Path
from collections import namedtuple

from ktpbase.database import DataBase
from ktpbase.log import logging

from openstf.feature_engineering.general import calc_completeness
from openstf.model.figure import (
    convert_to_base64_data_uri,
    plot_data_series,
    plot_feature_importance,
)
from openstf.model.general import (
    ForecastType,
    pre_process_data,
    split_data_train_validation_test,
)
from openstf.model.prediction.creator import PredictionModelCreator
from openstf.model.trainer.creator import ModelTrainerCreator
from openstf.monitoring.teams import send_report_teams_better, send_report_teams_worse

# TODO make this config more central
# Set thresholds
COMPLETENESS_THRESHOLD = 0.5
MINIMAL_TABLE_LENGTH = 100
MAX_AGE_YOUNG_MODEL = 7
TRAINING_PERIOD_DAYS = 90  # Note, if hyperparam is available, this value is overwritten

split_model_data = namedtuple("model_data", ["train", "validation", "test"])


def is_data_sufficient(data):
    """Check if enough data is left after validation and cleaning to continue
        with model training.

    Args:
        data: pd.DataFrame() with cleaned input data.

    Returns:
        (bool): True if amount of data is sufficient, False otherwise.

    """
    logger = logging.get_logger(__name__)
    # Set output variable
    is_sufficient = True

    # Calculate completeness
    completeness = calc_completeness(data, time_delayed=True, homogenise=False)
    table_length = data.shape[0]

    # Check if completeness is up to the standards
    if completeness < COMPLETENESS_THRESHOLD:
        logger.warning(
            "Input data is not sufficient, completeness too low",
            completeness=completeness,
            completeness_threshold=COMPLETENESS_THRESHOLD,
        )
        is_sufficient = False

    # Check if absolute amount of rows is sufficient
    if table_length < MINIMAL_TABLE_LENGTH:
        logger.warning(
            "Input data is not sufficient, table length too short",
            table_length=table_length,
            table_length_threshold=MINIMAL_TABLE_LENGTH,
        )
        is_sufficient = False

    return is_sufficient


def train_model_pipeline(pj, context, retrain_young_models=False, compare_to_old=True):
    """Main function that controls the training process of one prediction job.

    On a high level: first, the input data is collected, features are applied and the
    model is trained and stored. Additionally models for energy components are also
    trained if requested. Model is saved if better than the old model unless specified
    otherwise.

    Args:
        pj (dict): Prediction job
        context (openstf.task.utils.TaskContext): Task context for logging
        retrain_young_models (bool): Retrain models younger then 7 days if True.
            Defaults to False.
        compare_to_old (bool): Only save new model when its better than the old model.
            Defaults to True.
    """

    # TODO Maybe this whole function can be replaced with a scikit learn pipeline?
    model_trainer = create_model_trainer(pj, context, retrain_young_models)

    if model_trainer is None:
        # previous young model is not retrained, end function
        return

    context.perf_meter.checkpoint("model-creation")

    split_input_data = preprocess_for_model_training(pj, context)

    context.perf_meter.checkpoint("preprocessing")

    train_model(model_trainer, split_input_data)

    context.perf_meter.checkpoint("training")

    split_predicted_data = predict_after_model_training(
        pj,
        model_trainer,
        split_input_data,
    )

    context.perf_meter.checkpoint("predicting")

    new_model_is_better = is_new_model_better(
        pj,
        context,
        model_trainer,
        split_input_data,
        compare_to_old,
    )

    context.perf_meter.checkpoint("evaluation")

    path_to_save = context.config.paths.trained_models
    write_results_new_model(
        pj,
        model_trainer,
        new_model_is_better,
        split_input_data,
        split_predicted_data,
        path_to_save,
    )

    send_report_teams(pj, model_trainer, new_model_is_better)

    context.perf_meter.checkpoint("writing results")


def create_model_trainer(pj, context, retrain_young_models=False):
    """Create model trainer and set hyperparams.

    Args:
        pj (dict): Prediction job
        context (openstf.task.utils.TaskContext): Task context for logging
        retrain_young_models (bool): Retrain models younger then 7 days if True.
            Defaults to False.

    Returns:
        Implementation of the AbstractModelTrainer class tailored to
            a specific algorithm.
    """
    # Make model trainer creator
    mc = ModelTrainerCreator(pj)

    # Make model trainer
    model_trainer = mc.create_model_trainer()

    # if model is too young AND retrain_young_models it not set, skip training
    if model_trainer.old_model_age < MAX_AGE_YOUNG_MODEL:
        if not retrain_young_models:
            context.logger.info(
                "Skip training of model",
                model_age_days=model_trainer.old_model_age,
                retrain_young_models=retrain_young_models,
            )
            return

    # Set optimized hyper parameters if available
    model_trainer.hyper_parameters.update(context.database.get_hyper_params(pj))

    return model_trainer


def preprocess_for_model_training(pj, context):
    """Pre-process model data. Clean data and apply features

    Args:
        pj (dict): Prediction job
        context (openstf.task.utils.TaskContext): Task context for logging

    Raises:
        ValueError: when data quality is insufficient

    Returns:
        dict: dict of cleaned data with features, split in train, validation and test
    """
    # Specify training period
    # use hyperparam training_period_days if available
    hyperparams = {"training_period_days": TRAINING_PERIOD_DAYS, "featureset_name": "D"}
    hyperparams.update(context.database.get_hyper_params())

    featureset = context.database.get_featureset(hyperparams["featureset_name"])
    datetime_start = datetime.utcnow() - timedelta(
        days=hyperparams["training_period_days"]
    )
    datetime_end = datetime.utcnow()

    # Get data from database
    data = context.database.get_model_input(
        pid=pj["id"],
        location=[pj["lat"], pj["lon"]],
        datetime_start=datetime_start,
        datetime_end=datetime_end,
    )

    # Pre-process data
    clean_data_with_features = pre_process_data(data, featureset)

    # Check if we have enough data left to continue
    if is_data_sufficient(clean_data_with_features) is False:
        raise ValueError("Input data quality insufficient, aborting!")

    # Split data in train, test and validation sets
    train_data, validation_data, test_data = split_data_train_validation_test(
        clean_data_with_features
    )
    return split_model_data(
        train=train_data,
        validation=validation_data,
        test=test_data,
    )


def train_model(model_trainer, split_input_data):
    """Train model.

    Args:
        model_trainer (): Implementation of the AbstractModelTrainer class tailored to
            a specific algorithm.
        split_input_data (dict): dict of cleaned data with features, split in train,
            validation and test
    """
    # Train model
    model_trainer.train(split_input_data.train_data, split_input_data.validation_data)


def predict_after_model_training(pj, model_trainer, split_input_data):
    """Predict data with new trained model.

    Args:
        pj (dict): Prediction job
        model_trainer (): Implementation of the AbstractModelTrainer class tailored to
            a specific algorithm.
        split_input_data (dict): dict of cleaned data with features, split in train,
            validation and test

    Returns:
        dict: dict of predicted data, split in train, validation and test
    """
    # Build predictor and create predictions
    predictor = PredictionModelCreator.create_prediction_model(
        pj,
        ForecastType.DEMAND,
        model_trainer.trained_model,
        model_trainer.confidence_interval,
    )
    train_predict = predictor.make_forecast(
        split_input_data.train.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore")
    )
    validation_predict = predictor.make_forecast(
        split_input_data.validation.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore")
    )
    test_predict = predictor.make_forecast(
        split_input_data.test.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore")
    )

    return split_model_data(
        train=train_predict,
        validation=validation_predict,
        test=test_predict,
    )


def is_new_model_better(
    pj, context, model_trainer, split_input_data, compare_to_old=True
):
    """Evaluate quality of new model with old model and store it if it is better.

    Model is saved if better than the old model unless specified
    otherwise.

    Args:
        pj (dict): Prediction job
        context (openstf.task.utils.TaskContext): Task context for logging
        model_trainer (MLModelType): model trainer
        split_input_data (dict): dict of cleaned data with features, split in train,
            validation and test
        compare_to_old (bool, optional): Compare new trained model to old trained model
            if True. Defaults to True.
    """

    # Combine validation + train data
    combined = split_input_data.train.append(split_input_data.validation)

    # Evaluate model and store if better than old model
    # TODO Use predicted data series created above instead of predicting again
    # in better_than_old_model.
    if not compare_to_old or model_trainer.better_than_old_model(combined):
        context.logger.info("New model is better then old model", prediction_job=pj)
        model_trainer.store_model()
        new_model_is_better = True
    else:
        context.logger.warning("Old model is better then new model", prediction_job=pj)
        new_model_is_better = False
    return new_model_is_better


def create_evaluation_figures(model_trainer, split_input_data, split_predicted_data):
    """Create figures of feature importance and data series of input and predicted data.

    Args:
        model_trainer (MLModelType): model trainer
        split_input_data (dict): dict of cleaned data with features, split in train,
            validation and test
        split_predicted_data (dict): dict of predicted data, split in train, validation
            and test

    Returns:
        tuple: tuple containing:
            plotly.graph_objects.Figure: A treemap of the features.
            dict: dict of line plots of of split input and predicted
                data series per prediction horizon.
    """

    # Create figures
    figure_features = plot_feature_importance(model_trainer.feature_importance)
    horizons = split_input_data.train.Horizon.unique()
    input_data_list = [
        split_input_data.train,
        split_input_data.validation,
        split_input_data.test,
    ]
    predicted_data_list = [
        split_predicted_data.train,
        split_predicted_data.validation,
        split_predicted_data.test,
    ]
    figure_series = {
        f"Predictor{horizon}": plot_data_series(
            data=input_data_list,
            predict_data=predicted_data_list,
            horizon=horizon,
        )
        for horizon in horizons
    }
    return figure_features, figure_series


def write_results_new_model(
    pj,
    model_trainer,
    new_model_is_better,
    split_input_data,
    split_predicted_data,
    path_to_save,
):
    """Save results of new model to disk.

    Args:
        pj (dict): Prediction job
        model_trainer (MLModelType): model trainer
        new_model_is_better (boolean): True if new model is better than old model
        split_input_data (dict): dict of cleaned data with features, split in train,
            validation and test
        split_predicted_data (dict): dict of predicted data, split in train, validation
            and test
        path_to_save (pathlib.Path): path to save figures of results of new model
    """
    figure_features, figure_series = create_evaluation_figures(
        model_trainer, split_input_data, split_predicted_data
    )
    if new_model_is_better:
        # Save figures to disk
        save_loc = Path(path_to_save) / str(pj["id"])
    else:
        # Save figures to disk
        save_loc = Path(path_to_save) / str(pj["id"]) / "worse_model"

    os.makedirs(save_loc, exist_ok=True)
    figure_features.write_html(str(save_loc / "weight_plot.html"))
    for key, fig in figure_series.items():
        fig.write_html(str(save_loc / f"{key}.html"), auto_open=False)
        if not new_model_is_better:
            # write for a worse new model the series figures
            fig.update_layout(
                margin=dict(l=10, r=20, t=35, b=0),
                title=None,
                showlegend=False,
                grid=None,
                meta=None,
                xaxis=dict(visible=False),
                yaxis=dict(visible=False),
            )
            fig.write_image(save_loc / "f{key}.jpg", format="jpg", scale=0.3)
            convert_to_base64_data_uri(
                save_loc / "f{key}.jpg",
                save_loc / "f{key}.datauri",
                content_type="image/jpg",
            )


def send_report_teams(pj, model_trainer, new_model_is_better):
    """Send Teams message about result of new model.

    Args:
        pj (dict): Prediction job
        model_trainer (MLModelType): model trainer
        new_model_is_better (boolean): True if new model is better than old model
    """
    if new_model_is_better:
        send_report_teams_better(pj, model_trainer.feature_importance)
    else:
        send_report_teams_worse(pj)


def train_specific_model(context, pid):
    """Train model for given prediction id.

    Tracy-compatible function to train a specific model based on the prediction id (pid)
    Should not be used outside of Tracy, preferred alternative:
        train_model_pipeline

    Args:
        pid (int): Prediction id of the corresponding prediction job.

    Returns:
        Trained model (FIXME can be various datatypes at present)
    """
    # Get DataBase instance:
    db = DataBase()

    # Get prediction job based on the given prediction ID (pid)
    pj = db.get_prediction_job(pid)

    # Train model for pj
    train_model_pipeline(pj, context, compare_to_old=False, retrain_young_models=True)



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime

import pandas as pd

from openstf.model.capacity.train import train_capacity_prognosis
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext


def main():
    with TaskContext(__file__) as context:
        # training horizons
        y_hor = [0, 6, 13]

        # define input range
        datetime_end = datetime.utcnow()
        datetime_start = datetime_end - pd.Timedelta("400D")

        model_type = ["xgb", "lgb"]
        PredictionJobLoop(context, model_type=model_type).map(
            train_capacity_prognosis,
            datetime_start=datetime_start,
            datetime_end=datetime_end,
            y_hor=y_hor,
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timedelta

from openstf.model.capacity.predict import predict_capacity_prognosis
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext


def main():
    with TaskContext(__file__) as context:
        # prediction horizons
        y_hor = list(range(13))

        # define input range
        datetime_start = datetime.utcnow().date() - timedelta(days=30)
        datetime_end = datetime.utcnow().date() + timedelta(days=max(y_hor) + 1)
        model_type = ["xgb", "lgb"]

        PredictionJobLoop(context, model_type=model_type).map(
            predict_capacity_prognosis,
            datetime_start=datetime_start,
            datetime_end=datetime_end,
            y_hor=y_hor,
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import argparse
import importlib
import pkgutil

from openstf import PROJECT_ROOT


def parse_cli_arguments():
    # create the top-level parser
    parser = argparse.ArgumentParser(prog="Short term forecasting")

    subparsers = parser.add_subparsers(help="Run a task")

    # create the parser for the "task" command
    parser_task = subparsers.add_parser("task")
    parser_task.add_argument(
        "name", action="store", type=str, help="Name of the task you want to run."
    )

    args = parser.parse_args()
    return args


def validate_task_name(task_name):
    task_pkg_dir = PROJECT_ROOT / "openstf" / "tasks"
    task_modules = [m for _, m, _ in pkgutil.iter_modules([task_pkg_dir])]

    if task_name not in task_modules:
        raise RuntimeError(
            f"Unknown task: '{task_name}', valid tasks are :{', '.join(task_modules)}."
        )


def main():
    args = parse_cli_arguments()

    task_name = args.name

    validate_task_name(task_name)

    # import task
    task = importlib.import_module(f"openstf.tasks.{task_name}")
    # run task
    task.main()


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# -*- coding: utf-8 -*-
"""create_basecase_forecast.py

This module should be executed once every day. For all prediction_jobs, it will
create a 'basecase' forecast which is less accurate, but (almost) always available.
For now, it uses the load a week earlier.
Missing datapoints are interpolated.

Example:
    This module is meant to be called directly from a CRON job. A description of the
    CRON job can be found in the /k8s/CronJobs folder.

    Alternatively this code can be run directly by running:

        $ python create_basecase_forecast.py
"""
import openstf.model.predict as predict
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext


def main():
    with TaskContext(__file__) as context:
        model_type = ["xgb", "lgb"]

        PredictionJobLoop(context, model_type=model_type).map(
            predict.make_basecase_prediction
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import timedelta

import numpy as np
import pandas as pd
import pytz
import lightgbm as lgb
from ktpbase.log import logging

from openstf.model import metrics
from openstf.model.general import (
    pre_process_data,
    remove_features_not_in_set,
    split_data_train_validation_test,
)
from openstf.model.trainer.trainer import AbstractModelTrainer

# Available trainings period durations for optimization
# After preprocessing, the data consists of 75 days (of the original 90 days).
# To prevent overfitting on a very short period of time, the test and validation
# sets are always a constant fraction of the total data period (75 days).
# We take an optimized fraction of the remaining data as our train-data.
# Given the current fractions for test and validation, the longest train-block
# is 60 days.


class LGBModelTrainer(AbstractModelTrainer):
    def __init__(self, pj):
        super().__init__(pj)
        self.logger = logging.get_logger(self.__class__.__name__)
        # See LGBoost Parameters
        # https://lightgbm.readthedocs.io/en/latest/Python-Intro.html#setting-parameters
        # look at the parameters for Tree booster
        self.hyper_parameters = {"max_depth": 8}
        # self.hyper_parameters = {
        #     # general model specific hyper paramater (NOT optimized)
        #     # "silent": 1,
        #     "disable_default_eval_metric": 1,
        #     # model specific hyper parameters (optimized)
        #     "subsample": 0.9,
        #     "min_child_weight": 4,
        #     "max_depth": 8,
        #     "gamma": 0.5,
        #     "colsample_bytree": 0.85,
        #     "eta": 0.1,
        #     # model specific hyper parameters (NOT optimized)
        #     "objective": "reg:squarederror",
        #     # generic hyper parameters
        #     "featureset_name": "G",
        #     "training_period_days": 90,
        # }
        # private key lists for filtering
        self._lgb_hyper_parameter_keys = [
            # "silent",
            # "disable_default_eval_metric",
            # "subsample",
            # "min_child_weight",
            "max_depth",
            # "gamma",
            # "colsample_bytree",
            # "eta",
            # "objective",
        ]

    @property
    def feature_importance(self):
        """Return feature importances and weights of trained model.

        Returns:
            pandas.DataFrame: A DataFrame describing the feature importances and
            weights of the trained model.

        """
        if self.trained_model is None:
            return None

        feature_gain = pd.DataFrame(
            self.trained_model.feature_importance(importance_type="gain"),
            index=["gain"],
        ).T
        feature_gain /= feature_gain.sum()

        feature_weight = pd.DataFrame(
            self.trained_model.feature_importance(importance_type="split"),
            index=["weight"],
        ).T
        feature_weight /= feature_weight.sum()

        feature_importance = pd.merge(
            feature_gain, feature_weight, left_index=True, right_index=True
        )
        feature_importance.sort_values(by="gain", ascending=False, inplace=True)

        return feature_importance

    def train(
        self, train_data, validation_data, callbacks=None, early_stopping_rounds=10
    ):
        """Method that trains LGBoost model based on train and validation data.

        Args:
            train_data (pandas.DataFrame): The train data. Assumed is that the
                first column is the label.
            validation_data (pandas.DataFrame): The validation data.
            callbacks (list of callable): List of callback functions that can be
                called at the end of each training iteration

        Returns:
            lightgbm.Booster: this object is also stored in the self.trained_model
            attribute.

        """
        # Convert train and validation sets to Dmatrix format for computational
        #  efficiency. Drop Horizon column
        dtrain = lgb.Dataset(
            train_data.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore"),
            label=train_data.iloc[:, 0],
        )  # [:,1:-1] excludes label and 'horizon' column
        dvalidation = lgb.Dataset(
            validation_data.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore"),
            label=validation_data.iloc[:, 0],
        )

        # Define data set to be monitored during training, the last(validation)
        #  will be used for early stopping

        # Define callback function that is used to calculate metrics while training
        def custom_eval(preds, dset):
            return "MAE", metrics.mae(dset.get_label(), preds), False

        # get the lgb (hyper) parameters
        params = {k: self.hyper_parameters[k] for k in self._lgb_hyper_parameter_keys}

        # Train and validate model
        self.trained_model = lgb.train(
            train_set=dtrain,
            params=params,
            valid_sets=[dtrain, dvalidation],
            valid_names=["train", "validation"],
            num_boost_round=500,  # Can be large because we are early stopping anyway
            feval=custom_eval,
            verbose_eval=False,
            early_stopping_rounds=early_stopping_rounds,
            callbacks=callbacks,
        )

        # Add confidence interval data
        self.calculate_confidence_interval(validation_data)

        # Return model object
        return self.trained_model

    def better_than_old_model(self, test_data, penalty_factor=1.2):
        """Method that checks if newly trained model is better than the old model.

        Args:
            test_data (pandas.DataFrame): Data to test the models on.
            penalty_factor (float): Factor used to penalize the old model.


        Returns:
            bool: True if the new model performs better, False otherwise.

        """
        # Check if the old model is not None and try to make a prediction with the old model
        if self.old_model is None:
            self.logger.warning("No old model available")
            return True
        else:
            try:
                # Ask old model for prediction
                prediction_old_model = self.old_model.predict(
                    lgb.Dataset(
                        test_data.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore")
                    )
                )
            except Exception as e:
                prediction_old_model = np.nan
                self.logger.error("Could not compare to old model:", str(e)[:20])
                return True

        # Check if returned object is not None and try to make a prediction with the new model
        if self.trained_model is None:
            self.logger.warning(
                "New model is not yet trained, could not compare performance!"
            )
            return False
        else:
            try:
                prediction_new_model = self.trained_model.predict(
                    lgb.Dataset(
                        test_data.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore")
                    ),
                    num_iteration=self.trained_model.best_iteration,
                )
            except Exception as e:
                self.logger.error(
                    "Could not get prediction from new model:", str(e)[:20]
                )
                return False

        # Calculate scores
        old_mae = metrics.mae(prediction_old_model, test_data.iloc[:, 0])
        new_mae = metrics.mae(prediction_new_model, test_data.iloc[:, 0])

        # Compare and return True if new model is better, False otherwise
        if new_mae < old_mae * penalty_factor or prediction_old_model is np.nan:
            return True

        return False

    def store_model(self):
        """Stores the model."""
        self._store_model()

    def calculate_confidence_interval(self, validation_data):
        """Gets confidence interval data.

        This data is required to add confidence intervals to the predictions.

        Args:
            validation_data (pandas.DataFrame): The data to get the confidence
            intervals over.

        Returns:
            pandas.DataFrame: The confidence interval data. This is also stored
            in the self.confidence_interval attribute.

        """

        # Define some variables
        predicted = None
        self.confidence_interval = pd.DataFrame()

        # Loop over Horizons and ask prediction for each specific horizon
        for horizon in validation_data.Horizon.unique():
            # Make subset for this specific horizon
            sub_val = validation_data[validation_data.Horizon == horizon]
            # Check if returned object is not None and try to make a prediction with the new model
            if self.trained_model is None:
                print("New model is not yet trained, could not compute corrections!")
            else:
                try:
                    predicted = self.trained_model.predict(
                        sub_val.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore"),
                        num_iteration=self.trained_model.best_iteration,
                    )
                except Exception as e:
                    print("Could not get prediction from new model:", str(e))

            # Calculate confidence interval for this horizon
            confidence_interval_horizon = self._calculate_confidence_interval(
                sub_val.iloc[:, 0], predicted
            )
            confidence_interval_horizon[
                "horizon"
            ] = horizon  # Label with respective horizon
            self.confidence_interval = self.confidence_interval.append(
                confidence_interval_horizon
            )

        return self.confidence_interval

    def training_period_objective(
        self,
        trial,
        error_function,
        unprocessed_data,
        training_durations_days,
        optimized_parameters,
        featuresets,
    ):
        """Objective function that picks the optimal training_duration value.

        This function should be used after the hyper_params_objective has been completed
        by optuna.

        Args:
            trial: optuna trial object that is passed on during hyper parameter
                optimalisation.
            error_function (callable): Function to calculate the error metric to be
                optimized, preferably one from openstf.model.metrics.
            unprocessed_data (pandas.DataFrame): Data and features that have not yet
                been pre-processed.
            training_durations_days (list of int): Candidate training durations.
            featuresets (dict): All feature sets: with keys the featureset_name and
                values a list of feature names.

        Returns:
            float: Error metric value that is either minimized or maximized by the
            optuna study.

        """

        # Make selection of training-duration
        training_period = trial.suggest_categorical(
            "training_period_days", training_durations_days
        )

        # Update hyper parameters
        self.hyper_parameters.update(optimized_parameters)

        # Set training duration and shrink incoming data accordingly
        datetime_start = (
            unprocessed_data.index.max() - timedelta(days=training_period)
        ).replace(tzinfo=pytz.UTC)
        shortened_data = unprocessed_data.loc[unprocessed_data.index > datetime_start]

        # Pre-process data
        clean_shortened_data_with_all_features = pre_process_data(shortened_data)

        # Apply optimized featureset
        featureset_name = optimized_parameters["featureset_name"]

        featureset = featuresets[featureset_name]
        total_data = remove_features_not_in_set(
            clean_shortened_data_with_all_features, featureset=featureset
        )

        # Split data in train, test and validation sets, note we are using the
        # backtest option here because we assume hyperparameters do not change
        # much over time
        train_data, validation_data, test_data = split_data_train_validation_test(
            total_data,
            test_fraction=0.1,
            validation_fraction=0.1,
            back_test=True,
        )

        # Train model
        model = self.train(train_data, validation_data)

        # Make prediction on test data and prepare dataframes for comparison
        prediction = pd.DataFrame(
            model.predict(
                lgb.Dataset(
                    test_data.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore")
                ),
                num_iteration=model.best_iteration,
            )
        )
        realised = pd.DataFrame(test_data.iloc[:, 0])
        prediction.index = realised.index  # Set correct DateTime index

        # Calculate error metric
        error = error_function(realised.iloc[:, 0], prediction.iloc[:, 0])
        return error

    def hyper_params_objective(
        self, trial, error_function, clean_data_with_all_features, featuresets
    ):
        """Objective function used during hyperparameter optimalization.

        This objective function picks hyperparameter values in the pre-defined
        (in this method) parameter space and returns a resulting error. This
        error metric is either maximized or minimized by optuna.
        The metric type can be chosen by passing an error_function.

        Args:
            trial: optuna trial object that is passed on during hyper parameter
                optimalisation.
            error_function (callable): Function to calculate the error metric to be
                optimized, preferably one from openstf.model.metrics.
            clean_data_with_all_features (pandas.DataFrame): Data and features ready for
                model training
            featuresets (dict): All feature sets: with keys the featureset_name and
                values a list of feature names.

        Returns:
            float: Error metric value that is either minimized or maximized by the
            optuna study.

        """

        # This is an example on how to implement training period optimalisation,
        # if this gets to large a separate objective function should be build
        # Configure parameter space
        # Here the six parameters that give the most benfefit to prediction quality are optimized:
        # 1. eta [xgb_default=0.3],
        #   - Analogous to learning rate in GBM,
        #   - Makes the model more robust by shrinking the weights on each step
        #   - Typical final values to be used: 0.01-0.2
        # 2. subsample [xgb_default=1]
        #   - Same as the subsample of GBM. Denotes the fraction of observations to be randomly sampled for each tree.
        #   - Lower values make the algorithm more conservative and prevents overfitting but too small values might lead to under-fitting.
        #   - Typical values: 0.5-1
        # 3. min_child_weight [xgb_default=1]
        #   - Defines the minimum sum of weights of all observations required in a child.
        #   - This is similar to min_child_leaf in GBM but not exactly. This refers to min “sum of weights” of observations while GBM has min “number of observations”.
        #   - Used to control over-fitting. Higher values prevent a model from learning relations which might be highly specific to the particular sample selected for a tree.
        #   - Too high values can lead to under-fitting hence.
        # 4. max_depth [xgb_default=6]
        #   - The maximum depth of a tree, same as GBM.
        #   - Used to control over-fitting as higher depth will allow model to learn relations very specific to a particular sample.
        #   - Typical values: 3-10
        # 5. gamma [xgb_default=0]
        #   - A node is split only when the resulting split gives a positive reduction in the loss function. Gamma specifies the minimum loss reduction required to make a split.
        #   - Makes the algorithm conservative. The values can vary depending on the loss function and should be tuned.
        # 6. colsample_bytree [xgb_default=1]
        #   - Similar to max_features in GBM. Denotes the fraction of columns to be randomly samples for each tree.
        #   - Typical values: 0.5-1
        #
        # More information about these and other parameters and their limits can be found at
        # https://www.analyticsvidhya.com/blog/2016/03/complete-guide-parameter-tuning-xgboost-with-codes-python/
        parameter_space = {
            "eta": trial.suggest_float("eta", 0.01, 0.2),
            "subsample": trial.suggest_float("subsample", 0.5, 1.0),
            "min_child_weight": trial.suggest_int("min_child_weight", 1, 6),
            "max_depth": trial.suggest_int("max_depth", 3, 10),
            "gamma": trial.suggest_float("gamma", 0.0, 1.0),
            "colsample_bytree": trial.suggest_float("colsample_bytree", 0.5, 1.0),
        }

        # TODO featureset and trial period optimization is probably general
        # for all model types. When we start using other models we might want to
        # move this to the base class
        featureset_names = list(featuresets.keys())
        featureset_name = trial.suggest_categorical("featureset_name", featureset_names)
        self.logger.debug(
            "Current iteration of model trainer",
            featureset_name=featureset_name,
            parameter_space=parameter_space,
        )
        # Update hyper parameters (used by self.train())
        self.hyper_parameters.update(parameter_space)
        self.hyper_parameters["featureset_name"] = featureset_name

        # remove features
        featureset = featuresets[featureset_name]
        clean_data_with_features = remove_features_not_in_set(
            clean_data_with_all_features, featureset
        )

        # Split data in train, test and validation sets, note we are using the
        # backtest option here because we assume hyperparameters do not change
        # much over time
        train_data, validation_data, test_data = split_data_train_validation_test(
            clean_data_with_features,
            test_fraction=0.1,
            validation_fraction=0.1,
            back_test=True,
        )
        # Train model
        model = self.train(train_data, validation_data)

        # Make prediction on test data and prepare dataframes for comparison
        prediction = pd.DataFrame(
            model.predict(
                lgb.Dataset(
                    test_data.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore")
                ),
                num_iteration=model.best_iteration,
            )
        )
        realised = pd.DataFrame(test_data.iloc[:, 0])
        prediction.index = realised.index  # Set correct DateTime index

        # Calculate error metric
        error = error_function(realised.iloc[:, 0], prediction.iloc[:, 0])
        return error



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import os
import pickle
from pathlib import Path

import numpy as np
import pandas as pd
import sklearn
import xgboost as xgb


class CapacityPredictionModel:
    def __init__(self, classes=None, hyper_params=None):
        """ set default hyper-parameters"""
        if hyper_params is None:
            self.hyper_params = {
                "objective": "multi:softprob",
                "learning_rate": 0.01,
                "col_sample_bytree": 0.85,
                "max_depth": 3,
                "n_estimators": 256,
                "verbosity": 0,
            }

        # definition of classification classes
        self.classes = classes

        # create xgb model
        self.model = xgb.XGBClassifier(kwargs=self.hyper_params)

    def train(self, train_x, train_y, val_x, val_y):
        """ train model """
        self.model.fit(
            train_x,
            train_y,
            eval_set=[(train_x, train_y), (val_x, val_y)],
            verbose=False,
        )

    def predict(self, x):
        # find best iteration (on validation set)
        best_iter = int(
            np.argmin(self.model.evals_result()["validation_1"]["mlogloss"])
        )

        # predict classes
        y_pred = self.model.predict(x, ntree_limit=best_iter)
        y_pred = pd.DataFrame(y_pred.flatten(), index=x.index)[0]

        # predict probabilities
        y_pred_prob = self.model.predict_proba(x)
        y_pred_prob = pd.DataFrame(y_pred_prob, index=x.index)

        return y_pred, y_pred_prob

    def evaluate(self, x, y_true):
        scores = {}

        # predict on x
        y_pred, _ = self.predict(x)

        # compute f1 score
        scores["f1"] = sklearn.metrics.f1_score(
            y_true.values,
            y_pred.values.flatten(),
            average="weighted",
            labels=np.unique(y_pred.values.flatten()),
        )

        # compute accuracy score
        scores["accuracy"] = sklearn.metrics.accuracy_score(
            y_true.values, y_pred.values.flatten()
        )

        return scores

    def save(self, directory):
        # check directory
        if not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

        # save model
        with open(directory / "model.pkl", "wb") as fh:
            pickle.dump(self.model, fh)

        # save classes
        with open(directory / "classes.pkl", "wb") as fh:
            pickle.dump(self.classes, fh)

    def load(self, directory):
        # load model
        model_file = Path(directory) / "model.pkl"
        with open(model_file, "rb") as fh:
            self.model = pickle.load(fh)

        # load classes
        classes_file = Path(directory) / "classes.pkl"
        with open(classes_file, "rb") as fh:
            self.classes = pickle.load(fh)



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

"""calculate_kpi.py
This module contains the CRON job that is periodically executed to calculate key
performance indicators (KPIs) and save them to the database.
This code assumes prognoses are available from the persistent storage. If these are not
available run create_forecast.py to train all models.

The folowing tasks are caried out:
  1: Calculate the KPI for a given pid. Ignore SplitEnergy
  2: Create figures
  3: Write KPI to database

Example:
    This module is meant to be called directly from a CRON job.
    Alternatively this code can be run directly by running::
        $ python calculate_kpi.py
Attributes:
"""
# Import builtins
from datetime import datetime, timedelta

# Import project modules
from openstf.model.performance import calc_kpi_for_specific_pid
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext

# Thresholds for retraining and optimizing
THRESHOLD_RETRAINING = 0.25
THRESHOLD_OPTIMIZING = 0.50


def check_kpi_pj(pj, context, start_time, end_time):
    kpis = calc_kpi_for_specific_pid(pj["id"], start_time=start_time, end_time=end_time)
    # Write KPI's to database
    context.database.write_kpi(pj, kpis)

    # Add pid to the list of pids that should be retrained or optimized if
    # performance is insufficient
    if kpis["47.0h"]["rMAE"] > THRESHOLD_RETRAINING:
        context.logger.warning(
            "Need to retrain model, retraining threshold rMAE 47h exceeded",
            t_ahead="47.0h",
            rMAE=kpis["47.0h"]["rMAE"],
            retraining_threshold=THRESHOLD_RETRAINING,
        )
        function_name = "train_specific_model"
        context.logger.info("Adding tracy job", function=function_name)
        context.database.ktp_api.add_tracy_job(pj["id"], function=function_name)

    if kpis["47.0h"]["rMAE"] > THRESHOLD_OPTIMIZING:
        context.logger.warning(
            "Need to optimize hyperparameters, optimizing threshold rMAE 47h exceeded",
            t_ahead="47.0h",
            rMAE=kpis["47.0h"]["rMAE"],
            optimizing_threshold=THRESHOLD_OPTIMIZING,
        )
        function_name = "optimize_hyperparameters_for_specific_pid"
        context.logger.info("Adding tracy job", function=function_name)
        context.database.ktp_api.add_tracy_job(pj["id"], function=function_name)


def main():
    with TaskContext(__file__) as context:
        model_type = ["xgb", "lgb"]
        # Set start and end time
        start_time = datetime.date(datetime.utcnow()) - timedelta(days=1)
        end_time = datetime.date(datetime.utcnow())

        if datetime.utcnow().weekday() in [0, 1, 2, 3, 4]:
            PredictionJobLoop(context, model_type=model_type).map(
                check_kpi_pj,
                context,
                start_time=start_time,
                end_time=end_time,
            )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timedelta

import optuna
from ktpbase.database import DataBase
from ktpbase.log import logging

from openstf.model import metrics
from openstf.model.general import pre_process_data
from openstf.model.train import is_data_sufficient
from openstf.model.trainer.creator import ModelTrainerCreator
from openstf.monitoring.teams import post_teams

# Available trainings period durations for optimization
TRAINING_DURATIONS_DAYS = [30, 60, 90, 120]
MAX_AGE_HYPER_PARAMS_DAYS = 31
SHOW_OPTUNA_PROGRESS_BAR = False


def optimize_hyperparameters_pj(pj):
    """Optimized hyperparameters for a specific prediction job.
    First the age of the last stored hyperparameters is determined. If these are older
     than MAX_AGE_HYPER_PARAMS_DAYS, the hyper parameters will be optimized.

    Args:
        pj: (dict) Prediction job
    """

    db = DataBase()

    # Get data of last stored hyper parameters
    last = db.get_hyper_params_last_optimized(pj)

    # Check if hyper parameters are older than limit
    limit = datetime.utcnow() - timedelta(days=MAX_AGE_HYPER_PARAMS_DAYS)
    if last < limit:
        optimize_hyperparameters(pj["id"])


def optimize_hyperparameters(pid, n_trials=150, datetime_end=datetime.utcnow()):
    """Optimize the hyper parameters for a given prediction job.

        This function optimizes the model specific hyperparameters,
        featureset and training period. After optimization the hyperparameters
        are stored in the database and finally a teams messsage with the results
        is posted.

    Args:
        pid (int): Prediction job ID
        n_trials (int): Maximum number of trials/iteration first stage of
            hyper parameter optimization

    Returns:
        (dict): Dictionary with optimized hyperparameters, keys of the dictionarry
            depend on the model type.
    """

    db = DataBase()
    # initialize logging
    logger = logging.get_logger(__name__)

    max_training_period_days = max(TRAINING_DURATIONS_DAYS)
    # We do not want to train components during hyperparameter optimalisation
    train_components = 0

    # Get prediciton job
    pj = db.get_prediction_job(pid)

    pj["train_components"] = train_components

    # Specify training period
    datetime_start = datetime_end - timedelta(days=max_training_period_days)

    # Get data from database
    data = db.get_model_input(
        pid=pj["id"],
        location=[pj["lat"], pj["lon"]],
        datetime_start=datetime_start,
        datetime_end=datetime_end,
    )
    featuresets = db.get_featuresets()
    # Pre-process data
    clean_data_with_features = pre_process_data(data)

    # Initialize model trainer creator factory object
    mc = ModelTrainerCreator(pj)
    # Initialize model trainer
    model_trainer = mc.create_model_trainer()

    # Check if we have enough data left to continue
    if is_data_sufficient(clean_data_with_features) is False:
        raise ValueError("Input data quality insufficient, aborting!")

    # Setup optuna study
    optuna.logging.set_verbosity(optuna.logging.WARNING)
    # TODO make use of integrated XGBoost callbacks for efficient pruning and
    # faster optimalization runs, when bug in optana is solved
    first_stage_study = optuna.create_study(
        pruner=optuna.pruners.MedianPruner(n_warmup_steps=5), direction="minimize"
    )
    second_stage_study = optuna.create_study(
        pruner=optuna.pruners.MedianPruner(n_warmup_steps=5), direction="minimize"
    )
    # Carry out hyper-parameters and featureset optimization study
    first_stage_study.optimize(
        lambda trial: model_trainer.hyper_params_objective(
            trial,
            metrics.mae,
            clean_data_with_all_features=clean_data_with_features,
            featuresets=featuresets,
        ),
        n_trials=n_trials,
        show_progress_bar=SHOW_OPTUNA_PROGRESS_BAR,
    )
    optimized_error = first_stage_study.best_trial.value
    logger.info(
        f"Optimized error before training-period optimization: {optimized_error}",
        optimized_error=optimized_error,
    )

    # Carry out training-period optimization study
    second_stage_study.optimize(
        lambda trial: model_trainer.training_period_objective(
            trial,
            metrics.mae,
            unprocessed_data=data,
            training_durations_days=TRAINING_DURATIONS_DAYS,
            # The training-period optimization makes use of the previously optimized parameters
            optimized_parameters=first_stage_study.best_trial.params,
            featuresets=featuresets,
        ),
        n_trials=8,  # Ideally you'd test every duration only once but optuna does not work that way
        timeout=200,
        show_progress_bar=SHOW_OPTUNA_PROGRESS_BAR,
    )
    optimized_parameters = {
        **first_stage_study.best_trial.params,
        **second_stage_study.best_trial.params,
    }
    optimized_error = second_stage_study.best_trial.value
    logger.info("Optimized parameters", optimized_parameters=optimized_parameters)
    logger.info("Final optimized error", optimized_error=optimized_error)

    # Write best hyperparameters to database
    db.write_hyper_params(pj, optimized_parameters)
    # Compose message for teams (this is still cgb specific)
    msg = format_teams_message(pj, optimized_parameters)
    # Sent message to Teams
    post_teams(msg)

    return optimized_parameters


# NOTE perhaps formatting a message with paramters
# does not belong to the responsibility of this module
def format_teams_message(pj, hyper_params):
    color = "#046b00"  # green

    # format all hyper parameters using limited precision for floats
    # make keys bold (**key**)
    values = []
    for k, v in hyper_params.items():
        if type(v) is float:
            values.append(f"**{k}**: {v:0.3f}")
            continue
        values.append(f"**{k}**: {v}")
    # join all {hyper_parameter_name}: {value}  pairs with a new line
    text = "".join([f"* {v}   \n" for v in values])

    msg = {
        "fallback": "Optimized hyperparameters for model {0} {1}".format(
            pj["name"], pj["description"]
        ),
        "title": "Optimized hyperparameters for {0} {1}".format(
            pj["name"], pj["description"]
        ),
        "text": text,
        "color": color,
    }
    return msg



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timedelta, timezone

import pandas as pd
from ktpbase.database import DataBase
from ktpbase.log import logging

from openstf import feature_engineering
from openstf.model import validation_robot
from openstf.postproces import postprocess
from openstf.model.general import ForecastType
from openstf.model.prediction.creator import PredictionModelCreator

# configuration constants
FLATLINER_THRESHOLD = 6
COMPLETENESS_THRESHOLD = 0.7
FEATURES_H_AHEAD = 0.25
# cache input (for components predictions)
_input_data_cache = {}


def make_prediction(pj, forecast_type=ForecastType.DEMAND):
    logger = logging.get_logger(__name__)
    logger.info(
        "Start making prediction",
        prediction_id=pj["id"],
        customer_name=pj["name"],
        prediction_type="prediction",
    )
    # preparation ######################################################################
    forecast_start, forecast_end = generate_forecast_datetime_range(
        resolution_minutes=pj["resolution_minutes"],
        horizon_minutes=pj["horizon_minutes"],
    )
    datetime_start, datetime_end = generate_inputdata_datetime_range(
        t_behind_days=14, t_ahead_days=3
    )

    prediction_model = PredictionModelCreator.create_prediction_model(
        pj=pj, forecast_type=forecast_type
    )

    # load input data from database ###################################################
    input_data = get_model_input(
        pj=pj, datetime_start=datetime_start, datetime_end=datetime_end
    )
    # pre process input data ##########################################################
    preprocessed_input_data = pre_process_input_data(input_data, FLATLINER_THRESHOLD)

    # feature engineering #############################################################
    input_data_with_features = perform_feature_engineering(
        preprocessed_input_data, prediction_model.feature_names
    )
    # make forecast ###################################################################
    # Create correct format for to-be-forecasted times
    forecast_input_data = input_data_with_features.loc[
        forecast_start:forecast_end, prediction_model.feature_names
    ]
    completeness = prediction_model.calculate_completeness(forecast_input_data)
    # make forecast or fallback forecast
    if is_complete_enough(completeness, COMPLETENESS_THRESHOLD) is True:
        forecast = prediction_model.make_forecast(forecast_input_data)
    else:
        forecast = prediction_model.make_fallback_forecast(
            forecast_input_data=forecast_input_data, load_data=input_data[["load"]]
        )

    # post process forecast ###########################################################
    # Add wind and solar components
    forecast["forecast"] = postprocess.post_process_wind_solar(
        forecast["forecast"], forecast_type
    )

    # save forecast to database #######################################################
    DataBase().write_forecast(forecast, t_ahead_series=True)

    return forecast


def make_components_prediction(pj):
    logger = logging.get_logger(__name__)
    logger.info("Make components prediction", prediction_id=pj["id"])
    # select time period for the coming two days
    datetime_start, datetime_end = generate_inputdata_datetime_range(
        t_behind_days=0, t_ahead_days=3
    )
    logger.info(
        "Get predicted load", datetime_start=datetime_start, datetime_end=datetime_end
    )
    # Get most recent load forecast
    forecast = DataBase().get_predicted_load(
        pj, start_time=datetime_start, end_time=datetime_end
    )
    # Check if forecast is not empty
    if len(forecast) == 0:
        logger.warning(f'No forecast found. Skipping pid {pj["id"]}')
        return

    forecast["pid"] = pj["id"]
    forecast["customer"] = pj["name"]
    forecast["description"] = pj["description"]
    forecast["type"] = pj["typ"]
    forecast = forecast.drop(["stdev"], axis=1)

    logger.info("retrieving weather data")
    # Get required weather data
    weather_data = DataBase().get_weather_data(
        [pj["lat"], pj["lon"]],
        ["radiation", "windspeed_100m"],
        datetime_start=datetime_start,
        datetime_end=datetime_end,
        source="optimum",
    )

    # Get splitting coeficients
    split_coefs = DataBase().get_energy_split_coefs(pj)

    if len(split_coefs) == 0:
        logger.warning(f'No Coefs found. Skipping pid {forecast["pid"]}')
        return

    # Make component forecasts
    try:
        forecasts = postprocess.split_forecast_in_components(
            forecast, weather_data, split_coefs
        )
    except Exception as e:
        # In case something goes wrong we fall back on aan empty dataframe
        logger.warning(
            f"Could not make component forecasts: {e}, falling back on series of zeros!",
            exc_info=e,
        )
        forecasts = pd.DataFrame()

    # save forecast to database #######################################################
    DataBase().write_forecast(forecasts)
    logger.debug("Written forecast to database")

    return forecasts


def make_basecase_prediction(pj):
    logger = logging.get_logger(__name__)
    fill_limit = 4
    # preparation ######################################################################
    datetime_start, datetime_end = generate_inputdata_datetime_range(
        t_behind_days=15, t_ahead_days=0
    )
    logger.debug("Generated input timestamps")

    prediction_model = PredictionModelCreator.create_prediction_model(
        pj=pj, forecast_type=ForecastType.BASECASE
    )
    logger.debug("Loaded prediction model")
    forecast_resolution = f'{pj["resolution_minutes"]}T'

    # load input data from database ###################################################
    # Get historic load
    # TODO what to do when no load data is available
    historic_load = DataBase().get_load_pid(
        pid=pj["id"],
        datetime_start=datetime_start,
        datetime_end=datetime_end,
        forecast_resolution=forecast_resolution,
    )
    logger.debug("Retrieved historic load")

    # pre process input data ##########################################################

    # resample, fill missing values up to fill_limit
    historic_load = (
        historic_load.resample(forecast_resolution).mean().interpolate(limit=fill_limit)
    )
    # make forecast ###################################################################
    forecast = prediction_model.make_basecase_forecast(pj, historic_load)
    logger.debug("Made actual basecase prediction")
    # post process forecast ###########################################################

    # save forecast to database #######################################################
    DataBase().write_forecast(forecast, t_ahead_series=True)
    logger.debug("Written forecast to database")

    return forecast


# Helpers #############################################################################
# preparation
def generate_inputdata_datetime_range(t_behind_days=14, t_ahead_days=3):
    # get current date UTC
    date_today_utc = datetime.now(timezone.utc).date()
    # Date range for input data
    datetime_start = date_today_utc - timedelta(days=t_behind_days)
    datetime_end = date_today_utc + timedelta(days=t_ahead_days)

    return datetime_start, datetime_end


def generate_forecast_datetime_range(resolution_minutes, horizon_minutes):
    # get current date and time UTC
    datetime_utc = datetime.now(timezone.utc)
    # Datetime range for time interval to be predicted
    forecast_start = datetime_utc - timedelta(minutes=resolution_minutes)
    forecast_end = datetime_utc + timedelta(minutes=horizon_minutes)

    return forecast_start, forecast_end


# get data
def get_model_input(pj, datetime_start, datetime_end):
    # generate an unique identifier for storing/loading model input data
    # by doing so we don't have to retrieve the input data from the database
    # when making component predictions
    identifier = generate_input_data_id(pj, datetime_start, datetime_end)
    if identifier not in _input_data_cache.keys():
        input_data = DataBase().get_model_input(
            pid=pj["id"],
            location=(pj["lat"], pj["lon"]),
            # TODO it makes more sense to do the string converion in the
            # DataBase class and use pure Python datatypes in the rest of the code
            datetime_start=str(datetime_start),
            datetime_end=str(datetime_end),
        )
        # cache input data
        # TODO perhaps we can come up with something better? Or leave this to the
        # database. Perhaps its cheap anyway?
        _input_data_cache[identifier] = input_data
    else:
        input_data = _input_data_cache[identifier]

    return input_data


def generate_input_data_id(pj, datetime_start, datetime_end):
    return f'{pj["id"]}_{pj["lat"]}_{pj["lon"]}_{datetime_start}_{datetime_end}'


def _clear_input_data_cache():
    """Clear the input data cache dictionairy.

    This is mainly useful for testing.
    """
    global _input_data_cache
    _input_data_cache = {}


# pre processing
def pre_process_input_data(input_data, flatliner_threshold):
    logger = logging.get_logger(__name__)
    # Check for repeated load observations due to invalid measurements
    suspicious_moments = validation_robot.nonzero_flatliner(
        input_data, threshold=flatliner_threshold
    )

    if suspicious_moments is not None:
        # Covert repeated load observations to NaN values
        input_data = validation_robot.replace_invalid_data(
            input_data, suspicious_moments
        )
        # Calculate number of NaN values
        # TODO should this not be part of the replace_invalid_data function?
        num_nan = sum([True for i, row in input_data.iterrows() if all(row.isnull())])
        logger.warning(
            "Found suspicious data points, converted to NaN value",
            num_nan_values=num_nan,
        )

    return input_data


# feature engineering
def perform_feature_engineering(input_data, feature_names):
    minute_list = feature_engineering.general.extract_minute_features(
        feature_names=feature_names
    )

    input_data_with_features = feature_engineering.apply_features.apply_features(
        data=input_data, minute_list=minute_list, h_ahead=FEATURES_H_AHEAD
    )
    # Add missing features
    input_data_with_features = feature_engineering.general.add_missing_feature_columns(
        input_data=input_data_with_features, featurelist=feature_names
    )

    return input_data_with_features


# other
def is_complete_enough(completeness, completeness_threshold):
    logger = logging.get_logger(__name__)

    if completeness < completeness_threshold:
        logger.warning(
            "Forecast data completeness too low",
            completeness=completeness,
            completeness_threshold=completeness_threshold,
        )
        return False

    return True



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from openstf.model.prediction.xgboost.quantile import QuantileXGBPredictionModel
from openstf.model.prediction.xgboost.xgboost import XGBPredictionModel
from openstf.model.prediction.lightgbm.lightgbm import LGBPredictionModel


class PredictionModelCreator:

    PROGNOSIS_MODEL_CONSTRUCTORS = {
        "xgb": XGBPredictionModel,
        "xgb_quantile": QuantileXGBPredictionModel,
        "lgb": LGBPredictionModel,
    }

    @classmethod
    def create_prediction_model(cls, pj, forecast_type, model=None, confidence_df=None):

        model_type = pj["model"]

        if model_type not in cls.PROGNOSIS_MODEL_CONSTRUCTORS:
            raise KeyError(f"Unknown model_type: '{model_type}'")

        return cls.PROGNOSIS_MODEL_CONSTRUCTORS[model_type](
            pj, forecast_type, model, confidence_df
        )



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import traceback

from ktpbase.config.config import ConfigManager
from ktpbase.database import DataBase
from ktpbase.log import logging

from openstf.monitoring.teams import post_teams
from openstf.monitoring.performance_meter import PerformanceMeter
from openstf.tasks.utils.predictionjobloop import PredictionJobException
from openstf.tasks.utils import utils


class TaskContext:
    def __init__(
        self,
        task_file,
        suppress_exceptions=False,
        post_teams_on_exception=True,
        on_exception=None,
        on_successful=None,
        on_end=None,
    ):
        """A context manager that can be used to run tasks with.

        Should be used as:
        with TaskContext("__file__") as context:
            pass

        Args:
            task_file (string): Path of the task module provided by __file__
            suppress_exceptions (bool, optional): If set to False the context
                manager will pass any raised exception on. Defaults to False.
            post_teams_on_exception (bool, optional): If set to True the context
                manager will automatically post a message to teams when an
                exception is encountered. Defaults to True.
            on_exception (callable, optional): Callback, will be called
                when an exception is raised. Callable gets exc_type, exc_info,
                stack_info as arguments.
            on_successful (callable, optional): Callback, will be called
                everytime if the task is successful (no exception is raised).
            on_end (callable, optional): Callback, will be called if the task is
                completed. Callable gets a bool indicating success as argument.
        """
        self.name = utils.get_module_name(task_file)
        self.suppress_exceptions = suppress_exceptions
        self.post_teams_on_exception = post_teams_on_exception
        self.on_exception = on_exception
        self.on_successful = on_successful
        self.on_end = on_end

    def __enter__(self):
        self.config = ConfigManager.get_instance()

        logging.configure_logging(
            loglevel=self.config.loglevel,
            runtime_env=self.config.env,
        )
        self.logger = logging.get_logger(__name__).bind(task=self.name)

        self.database = DataBase()

        self.perf_meter = PerformanceMeter(self.logger)

        self.perf_meter.start_level("task", self.name)

        return self

    def __exit__(self, exc_type, exc_info, stack_info):
        self.successful = exc_type is None

        if self.successful:
            if self.on_successful is not None:
                self.on_successful()
        else:
            self.logger.error(
                "An error has occured",
                exc_info=exc_info,
                stack_info=stack_info,
            )
            if self.on_exception is not None:
                self.on_exception(exc_type, exc_info, stack_info)

            if self.post_teams_on_exception:
                self._send_teams_message(exc_type, exc_info, stack_info)

        if self.on_end is not None:
            self.on_end(self.successful)

        self.perf_meter.complete_level(self.successful)

        # Returning true stops the propagation of the exception
        return self.suppress_exceptions

    def _send_teams_message(self, exc_type, exc_info, stack_info):
        stack_text = "".join(traceback.format_exception(exc_type, exc_info, stack_info))
        # The teams webhook behaves weird. Using the correct order \r\n does NOT work
        # The 4 spaces forces teams to format a monospaced message for some reason
        stack_text = stack_text.replace("\n", "\n\r    ")
        msg = {
            "title": f"Task '{self.name}' raised an exception",
            "color": "#a54f4f",
            "sections": [],
        }
        if exc_type is PredictionJobException:
            metrics = exc_info.metrics
            msg["sections"].append(
                {
                    "facts": [
                        ("Number of pjs in this task", metrics["num_jobs"]),
                        ("Number of pjs started", metrics["jobs_started"]),
                        ("Number of pjs successful", metrics["jobs_successful"]),
                        ("Number of pjs unsuccessful", metrics["jobs_unsuccessful"]),
                    ],
                }
            )
            string_pids_unsuccessful = ", ".join(
                [f"{pid:d}" for pid in metrics["pids_unsuccessful"]]
            )
            msg["sections"].append(
                {
                    "facts": [
                        ("Unsuccessful pids", string_pids_unsuccessful),
                    ]
                }
            )

        msg["sections"].append(
            {
                "text": f"    {stack_text}",
            }
        )

        post_teams(msg)



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import timedelta
from enum import Enum

import numpy as np
import pandas as pd
from ktpbase.log import logging

from openstf import PROJECT_ROOT
from openstf.data_validation import data_validation
from openstf.feature_engineering.apply_features import apply_multiple_horizon_features
from openstf.feature_engineering.general import (
    apply_fit_insol,
    apply_persistence,
    remove_features_not_in_set,
)


# TODO replace this with ModelType (MLModelType == Machine Learning model type)
class MLModelType(Enum):
    XGB = "xgb"
    XGB_QUANTILE = "xgb_quantile"
    LGB = "lgb"


class ForecastType(Enum):
    DEMAND = "demand"
    WIND = "wind"
    SOLAR = "solar"
    BASECASE = "basecase"


# TODO move to config
PV_COEFS_FILEPATH = PROJECT_ROOT / "openstf" / "data" / "pv_single_coefs.csv"


def split_data_train_validation_test(
    data,
    test_fraction=0.0,
    validation_fraction=0.15,
    back_test=False,
    period_sampling=True,
    period_timedelta=timedelta(days=2),
):
    """Split input data into train, test and validation set.

    Function for splitting cleaned data with features in a train, test and
    validation dataset. In an operational setting the folowing sequence is
    returned:

    Test >> Validation >> Train

    For a back test (indicated with argument "back_test") the folowing sequence
    is returned:

    Validation >> Train >> Test

    The ratios of the different types can be set with test_fraction and
    validation fraction.

    Args:
        data (pandas.DataFrame): Clean data with features
        test_fraction (float): Number between 0 and 1 that indicates the desired
            fraction of test data.
        validation_fraction (float): Number between 0 and 1 that indicates the
            desired fraction of validation data.
        back_test (bool): Indicates if data is intended for a back test.
        period_sampling (bool): Indicates if validation data must be sampled as
            periods.
        period_timedelta (datetime.timedelta): Duration of the periods that will
            be sampled as validation data. Only used for period_sampling=True.

    Returns:
        Tuple with train data, validation data and test data:
            [0] (pandas.DataFrame): Train data
            [1] (pandas.DataFrame): Validation data
            [2] (pandas.DataFrame): Test data
    """
    MIN_TRAIN_FRACTION = 0.5
    logger = logging.get_logger(__name__)

    # Check input
    train_fraction = 1 - (test_fraction + validation_fraction)

    if train_fraction < 0:
        raise ValueError(
            "Test ({test_fraction}) and validation fraction ({validation_fraction}) too high."
        )

    if train_fraction < MIN_TRAIN_FRACTION:
        # TODO no action if above threshold? Which settings are meant here?
        logger.warning("Current settings only allow for 50% train data")

    # Get start date from the index
    start_date = data.index.min().to_pydatetime()

    # Calculate total of quarter hours (PTU's) in input data
    number_indices = len(data.index.unique())  # Total number of unique timepoints
    delta = (
        data.index.unique().sort_values()[1] - data.index.unique().sort_values()[0]
    )  # Delta t, assumed to be constant troughout DataFrame
    delta = timedelta(
        seconds=delta.seconds
    )  # Convert from pandas timedelta to original python timedelta

    # Default sampling, take a single validation set.
    if not period_sampling:
        # Define start and end datetimes of test, train, val sets based on input
        if back_test:
            start_date_val = start_date
            start_date_train = (
                start_date_val + np.round(number_indices * validation_fraction) * delta
            )
            start_date_test = (
                start_date_train
                + np.round(number_indices * (1 - validation_fraction - test_fraction))
                * delta
            )
            train_data = data[start_date_train:start_date_test]
            test_data = data[start_date_test:None]
        else:
            start_date_test = start_date
            start_date_val = (
                start_date + np.round(number_indices * test_fraction) * delta
            )
            start_date_train = (
                start_date_val + np.round(number_indices * validation_fraction) * delta
            )
            test_data = data[start_date_test:start_date_val]
            train_data = data[start_date_train:None]

        # In either case validation data is before the training data
        validation_data = data[start_date_val:start_date_train]

    # Sample periods in the training part as the validation set.
    else:
        if back_test:
            start_date_combined = start_date
            start_date_test = (
                start_date_combined
                + np.round(number_indices * (1 - test_fraction)) * delta
            )

            combined_data = data[start_date_combined:start_date_test]
            test_data = data[start_date_test:None]
        else:
            start_date_test = start_date
            start_date_combined = (
                start_date + np.round(number_indices * test_fraction) * delta
            )

            combined_data = data[start_date_combined:]
            test_data = data[start_date_test:start_date_combined]

        train_data, validation_data = sample_validation_data_periods(
            combined_data,
            validation_fraction=validation_fraction / (1 - test_fraction),
            period_length=int(period_timedelta / delta),
        )

    # Return datasets
    return train_data, validation_data, test_data


def sample_validation_data_periods(data, validation_fraction=0.15, period_length=192):
    """Splits data in train and validation dataset.

    Tries to sample random periods of given length to form a validation set of
    the desired size. Will raise an error if the number of attempts exceeds the
    maximum given to this function (default: 10).

    Args:
        data (pandas.DataFrame): Clean data with features
        validation_fraction (float): Number between 0 and 1 that indicates the
            desired fraction of validation data. Using a value larger than ~0.4 might
            lead to this function failing.
        period_length (int): Desired size of the sampled periods. The actual
            values can be slightly different if this is required to create the
            right fraction of validation data. Each period will have a duration
            of at least half period_length and at most one and a half
            period_length.

    Returns:
        train_data (pandas.DataFrame): Train data.
        validation_data (pandas.DataFrame): Validation data.
    """

    data_size = len(data.index.unique())
    validation_size = np.round(data_size * validation_fraction).astype(int)
    number_periods = np.round(validation_size / period_length).astype(int)

    # Always atleast one validation period
    if number_periods < 1:
        number_periods = 1

    period_lengths = [period_length] * (number_periods - 1)
    period_lengths += [validation_size - sum(period_lengths)]

    # Default buffer is equal to period_length
    buffer_length = period_length

    # Check if the dataset has enough points for the current settings
    if validation_size + 2 * number_periods * buffer_length >= data_size:
        # Use half period_length otherwise
        buffer_length = np.round(buffer_length / 2).astype(int)

    # Sample indices as validation data
    try:
        validation_indices = _sample_indices(
            data_size - max(period_lengths), period_lengths, buffer_length
        )
    except ValueError:
        raise ValueError(
            "Could not sample {} periods from data of size {}. Maybe the \
            validation_fraction is too high (>0.4)?".format(
                period_lengths, data_size
            )
        )

    # Select validation data
    validation_data = data.loc[data.index.unique()[validation_indices]]

    # Select the other data as training data
    train_data = data[~data.index.isin(validation_data.index)]

    return train_data, validation_data


def _sample_indices(number_indices, period_lengths, buffer_length):
    """Samples periods of given length with the given buffer.

    Args:
        number_indices (int): Total number of indices that are available for
            sampling.
        period_lengths (list:int): List of lengths for each period that will be
            sampled.
        buffer_length (int): Number of indices between each sampled period that
            will be removed from the sampling set.

    Returns:
        numpy.array: Sorted (ascending) list of sampled indices.

    """
    stockpile = set(range(number_indices))

    rng = np.random.default_rng()
    sampled = set()
    for period_length in period_lengths:
        # Sample random starting indices from indices set
        start_point = rng.choice(list(stockpile))
        end_point = start_point + period_length

        # Append sampled indices
        sampled |= set(range(start_point, end_point))

        # Remove sampled indices plus a buffer zone.
        stockpile -= set(
            range(
                start_point - period_length - buffer_length, end_point + buffer_length
            )
        )

    return np.sort(list(sampled))


def pre_process_data(data, featureset=None, horizons=None):
    """Function that automates the pre processing of the data.

    Args:
        data (pd.DataFrame): Data with (unvalidated) input data and without features.

    Returns:

        pd.DataFrame: Cleaned data with features.

    """
    if horizons is None:
        horizons = [0.25, 47]

    # Validate input data
    validated_data = data_validation.validate(data)

    # Apply features
    # TODO it would be nicer to only generate the required features
    validated_data_data_with_features = apply_multiple_horizon_features(
        validated_data, h_aheads=horizons
    )

    # remove features not in requested set if required
    if featureset is not None:
        validated_data_data_with_features = remove_features_not_in_set(
            validated_data_data_with_features, featureset
        )

    # Clean up data
    clean_data_with_features = data_validation.clean(validated_data_data_with_features)

    return clean_data_with_features


def combine_forecasts(forecasts, combination_coefs):
    """This function combines several independent forecasts into one, using
        predetermined coefficients.

    Input:
        - forecasts: pd.DataFrame(index = datetime, algorithm1, ..., algorithmn)
        - combinationcoefs: pd.DataFrame(param1, ..., paramn, algorithm1, ..., algorithmn)

    Output:
        - pd.DataFrame(datetime, forecast)"""

    models = [x for x in list(forecasts) if x not in ["created", "datetime"]]

    # Add subset parameters to df
    # Identify which parameters should be used to define subsets based on the
    # combinationcoefs
    subset_columns = [
        "tAhead",
        "hForecasted",
        "weekday",
        "hForecastedPer6h",
        "tAheadPer2h",
        "hCreated",
    ]
    subset_defs = [x for x in list(combination_coefs) if x in subset_columns]

    df = forecasts.copy()
    # Now add these subsetparams to df
    if "tAhead" in subset_defs:
        t_ahead = (df["datetime"] - df["created"]).dt.total_seconds() / 3600
        df["tAhead"] = t_ahead

    if "hForecasted" in subset_defs:
        df["hForecasted"] = df.datetime.dt.hour

    if "weekday" in subset_defs:
        df["weekday"] = df.datetime.dt.weekday

    if "hForecastedPer6h" in subset_defs:
        df["hForecastedPer6h"] = pd.to_numeric(
            np.floor(df.datetime.dt.hour / 6) * 6, downcast="integer"
        )

    if "tAheadPer2h" in subset_defs:
        df["tAheadPer2h"] = pd.to_numeric(
            np.floor((df.datetime - df.created).dt.total_seconds() / 60 / 60 / 2) * 2,
            downcast="integer",
        )

    if "hCreated" in subset_defs:
        df["hCreated"] = df.created.dt.hour

    # Start building combinationcoef dataframe that later will be multiplied with the
    # individual forecasts
    # This is the best way for a backtest:
    #    uniquevalues = list([np.unique(df[param].values) for param in subsetDefs])
    #    permutations = list(itertools.product(*uniquevalues))

    # This is the best way for a single forecast
    permutations = [tuple(x) for x in df[subset_defs].values]

    result_df = pd.DataFrame()

    for subsetvalues in permutations:
        subset = df.copy()
        coefs = combination_coefs

        # Create subset based on all subsetparams, for forecasts and coefs
        for value, param in zip(subsetvalues, subset_defs):
            subset = subset.loc[subset[param] == value]
            # Define function which find closest match of a value from an array of values.
            #  Use this later to find best coefficient from the given subsetting dividers
            closest_match = min(coefs[param], key=lambda x: abs(x - value))
            coefs = coefs.loc[coefs[param] == closest_match]
            # Find closest matching value for combinationCoefParams corresponding to
            # available subsetValues

        # Of course, not all possible subsets have to be defined in the forecast.
        # Skip empty subsets
        if len(subset) == 0:
            continue

        # Multiply forecasts with their coefficients
        result = np.multiply(subset[models], np.array(coefs[models]))
        result["forecast"] = result.apply(np.nansum, axis=1)
        # Add handling with NA values for a single forecast
        result["coefsum"] = np.nansum(coefs[models].values)
        nanselector = np.isnan(subset[models].iloc[0].values)
        result["nonnacoefsum"] = np.nansum(coefs[models].values.flatten() * nanselector)
        result["forecast"] = (
            result["forecast"]
            * result["coefsum"]
            / (result["coefsum"] - result["nonnacoefsum"])
        )
        result["datetime"] = subset["datetime"]
        result["created"] = subset["created"]
        result = result[["datetime", "created", "forecast"]]

        result_df = result_df.append(result)

    # for safety: remove duplicate results to prevent multiple forecasts for the same
    # time created
    # resultDF.drop_duplicates(keep='last', inplace = True)

    #    #rename created column to datetime and add datetime for export
    #    resultDF.reset_index(inplace = True)
    #    resultDF.columns = ['datetime','created', 'forecast']
    #
    # sort by datetime
    result_df.sort_values(["datetime", "created"], inplace=True)

    return result_df


def fides(data, all_forecasts=False):
    """Fides makes a forecast based on persistence and a direct fit with insolation.
    Input:
        - data = pd.DataFrame(index = datetime, columns =['output','insolation'])
    Optional:
        - allForecasts = Bool. Should all forecasts be returned or only the combination
    Example:
    import numpy as np
    index = pd.date_range(start = "2017-01-01 09:00:00", freq = '15T', periods = 300)
    data = pd.DataFrame(index = index,
                        data = dict(load=np.sin(index.hour/24*np.pi)*np.random.uniform(0.7,1.7, 300)))
    data['insolation'] = data.load * np.random.uniform(0.8, 1.2, len(index)) + 0.1
    data.loc[int(len(index)/3*2):,"load"] = np.NaN"""

    insolation_forecast = apply_fit_insol(data, add_to_df=False)
    persistence = apply_persistence(data, how="mean", smooth_entries=4, add_to_df=True)

    df = insolation_forecast.merge(persistence, left_index=True, right_index=True)

    coefs = pd.read_csv(PV_COEFS_FILEPATH)

    # Apply combination coefs
    df["created"] = df.loc[df.load.isnull()].index.min()
    forecast = combine_forecasts(
        df.loc[df.load.isnull(), ["forecaopenstfitInsol", "persistence", "created"]]
        .reset_index()
        .rename(columns=dict(index="datetime")),
        coefs,
    ).set_index("datetime")[["forecast"]]

    if all_forecasts:
        forecast = forecast.merge(
            df[["persistence", "forecaopenstfitInsol"]],
            left_index=True,
            right_index=True,
            how="left",
        )

    return forecast



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# -*- coding: utf-8 -*-
"""model_train.py

This module contains the CRON job that is periodically executed to retrain the
prognosis models. For this the folowing steps are caried out:
  1. Get historic training data (TDCV, Load, Weather and APX price data)
  2. Apply features
  3. Train and Test the new model
  4. Check if new model performs better than the old model
  5. Store the model if it performs better
  6. Send slack message to inform the users

Example:
    This module is meant to be called directly from a CRON job. A description of
    the CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::

        $ python model_train.py

"""
from openstf.model.train import train_model_pipeline
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext


def main():
    with TaskContext(__file__) as context:
        model_type = ["xgb", "xgb_quantile", "lgb"]

        PredictionJobLoop(context, model_type=model_type).map(
            train_model_pipeline, context
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import pandas as pd
import scipy
import lightgbm as lgb
from ktpbase.log import logging

import openstf.feature_engineering.general as feature_engineering
from openstf.model.prediction.prediction import AbstractPredictionModel


class LGBPredictionModel(AbstractPredictionModel):
    def __init__(
        self, pj, forecast_type, trained_model=None, trained_confidence_df=None
    ):
        super().__init__(pj, forecast_type, trained_model, trained_confidence_df)
        self.logger = logging.get_logger(self.__class__.__name__)

    @property
    def feature_names(self):
        return self.model.feature_names

    def calculate_completeness(self, forcast_input_data):
        """Calculate the completeness

        Args:
            forcast_input_data (pandas.DataFrame): Input data

        Returns:
            float: The completeness of the input data
        """

        # Check if forcast_input_data is 'complete' enough.
        scores = self.model.feature_importance(importance_type="gain")

        for feature in self.model.feature_names:
            if feature not in scores.keys():
                scores.update({feature: 0})

        weights = pd.DataFrame(index=[0], data=scores)
        weights = weights.loc[:, list(forcast_input_data.columns)]

        completeness = feature_engineering.calc_completeness(
            forcast_input_data, weights.values[0]
        )

        self.logger.info(f"Completeness: {completeness:.2f}")

        return completeness

    def make_forecast(self, forcast_input_data):
        """Make a forecast using a trained model

        Args:
            forcast_input_data (pandas.DataFrame): Forcast (section) input data

        Returns:
            pandas.DataFrame: The forecase
        """
        self.logger.info("Use LightGBM model to make a forecast")

        forecast_request = lgb.Dataset(forcast_input_data)

        model_forecast = self.model.predict(
            forecast_request, num_iteration=self.model.best_iteration
        )
        forecast = pd.DataFrame(
            index=forcast_input_data.index, data={"forecast": model_forecast}
        )

        # Add standard deviation from `confidence_df` or file (if exists and is not empty)
        forecast = self._add_standard_deviation_to_forecast(forecast)

        forecast = self._add_quantiles_to_forecast(forecast, self.pj["quantiles"])

        # Add prediction job property to forecast
        forecast = self.add_prediction_job_properties_to_forecast(
            pj=self.pj,
            forecast=forecast,
            algorithm_type=str(self.model_file_path),
            forecast_type=self.forecast_type,
        )

        npoints = len(forecast)

        self.logger.info(f"Made a forecast, npoints: {npoints}", npoints=npoints)

        return forecast

    @staticmethod
    def _add_quantiles_to_forecast(forecast, quantiles):
        """Add quantiles to forecast.

        Use the standard deviation to calculate the quantiles.

        Args:
            forecast (pd.DataFrame): Forecast (should contain a 'forecast' + 'stdev' column)
            quantiles (list): List with desired quantiles

        Returns:
            (pd.DataFrame): Forecast DataFrame with quantile (e.g. 'quantile_PXX')
                columns added.

        """

        # Check if stdev and forecast are in the dataframe
        if not all(elem in forecast.columns for elem in ["forecast", "stdev"]):
            raise ValueError("Forecast should contain a 'forecast' and 'stdev' column")

        for quantile in quantiles:
            quantile_key = f"quantile_P{quantile * 100:02.0f}"
            forecast[quantile_key] = (
                forecast["forecast"]
                + scipy.stats.norm.ppf(quantile) * forecast["stdev"]
            )

        return forecast
